/******************************************************************************
* Netzwerk-Exporte bearbeiten																   *
* 																										*
******************************************************************************/

/* Die Includes werden gemacht */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <support.h>
#include <string.h>
#include <ctype.h>
#include "util.h"



/*Lokale Defines */
/*---------------------------------------------------------------------------*/

/*in print-error wird eine Fehlerseite generiert und ausgegeben*/
/*---------------------------------------------------------------------------*/
void print_error(char* error){

	header(NONE,"Fehler");
	printf("%s\n",error);
	tail();
}


/*in add_export wird ein Import in fstab eingetragen*/
/*---------------------------------------------------------------------------*/
int add_export(struct structExport *export_base,
					char smb_name[MAXLINE],
             char	path[MAXLINE],
             char type[32],
             int	read_only,
             int	nfs_root_access) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structExport *akt;
	struct structExport *base;
	char					*cmdstring;
	int					smb_export;
	int					nfs_export;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   
   if (strcmp(smb_name,""))
   	if (smbNameExists(export_base, smb_name)) return SMB_NAME_EXISTS;
   
   if (ExportExists(export_base, path)) return EXPORT_EXISTS;

   if (!strcmp(type,"smb und nfs")) {
   	nfs_export=TRUE;
   	smb_export=TRUE;
   }
   if (!strcmp(type,"default")) {
   	nfs_export=TRUE;
   	smb_export=TRUE;
   }
   if (!strcmp(type,"smb")) {
   	smb_export=TRUE;
   	nfs_export=FALSE;
   }
   if (!strcmp(type,"nfs")) {
   	nfs_export=TRUE;
   	smb_export=FALSE;
   }
	
	if (!strcmp(smb_name,"") && smb_export) return SMB_NAME_NOT_SET;
	
	if (DEBUG) {
		printf("add_export(): Kontrollausgabe<br>\n");
		akt = export_base;
		if (export_base != NULL) do {
			printf("add_export():Pfad: %s NFS: %d SMB: %d <br>\n",akt->path,akt->nfs_export,akt->smb_export);
			akt = akt->next;
		} while (akt != export_base);
	}

	akt = (struct structExport *) malloc(sizeof(struct structExport));
	strcpy(akt->smb_name,smb_name);
	strcpy(akt->path,path);
	strcpy(akt->nfs_unknown_options,"");
	strcpy(akt->nfs_common_unknown_options,"");
	strcpy(akt->smb_unknown_options,"");
	akt->smb_export=smb_export;   
	akt->nfs_export=nfs_export;
	akt->nfs_root_access=nfs_root_access;
	akt->read_only=read_only;
	akt->public=TRUE;
	akt->printable=FALSE;
	addExport(&export_base,akt);

	if (DEBUG) {
		printf("add_export(): Kontrollausgabe<br>\n");
		akt = export_base;
		if (export_base != NULL) do {
			printf("add_export():Pfad: %s NFS: %d SMB: %d <br>\n",akt->path,akt->nfs_export,akt->smb_export);
			akt = akt->next;
		} while (akt != export_base);
	}
	
	if (!FlushExports(export_base)) {
		system(NFS_RESTART_CMD);
		system(SMB_RESTART_CMD);
	}
	return 0;
}


/*in delete_export wird ein Import aus der fstab geloescht*/
/*---------------------------------------------------------------------------*/
int delete_export(struct structExport *export_base,
					char smb_name[MAXLINE],
             char	path[MAXLINE],
             char type[32],
             int	read_only,
             int	nfs_root_access) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structExport *akt;
	char					*cmdstring;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   
   if (!ExportExists(export_base,path)) return EXPORT_UNKNOWN;
   akt=GetExport(export_base,path);
   delExport(&export_base,path); 
	if (!FlushExports(export_base)) {
		system(NFS_RESTART_CMD);
		system(SMB_RESTART_CMD);
	}
	return 0;
}


/*mit change_export koennen die mount-Parameter veraendert werden*/
/*---------------------------------------------------------------------------*/
int change_export(struct structExport *export_base,
					char smb_name[MAXLINE],
             char	path[MAXLINE],
             char type[32],
             int	read_only,
             int	nfs_root_access) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structExport *akt;
	char					*cmdstring;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   if (!ExportExists(export_base,path)) return EXPORT_UNKNOWN;
   akt=GetExport(export_base,path);
   if (strcmp(smb_name,"")) strcpy(akt->smb_name,smb_name);
   if (strcmp(type,"default")) {
	   if (!strcmp(type,"smb und nfs")) {
	   	akt->nfs_export=TRUE;
	   	akt->smb_export=TRUE;
	   }
	   if (!strcmp(type,"smb")) {
	   	akt->smb_export=TRUE;
	   	akt->nfs_export=FALSE;
	   }
	   if (!strcmp(type,"nfs")) {
	   	akt->nfs_export=TRUE;
  		 	akt->smb_export=FALSE;
  		}
   }
	akt->read_only=read_only;
	akt->nfs_root_access=nfs_root_access;
	if (DEBUG) printf("change_export: Pfad: %s, read_only: %d, nfs_root_access: %d<br>\n",akt->path,akt->read_only,akt->nfs_root_access);
	if (!FlushExports(export_base)) {
		system(NFS_RESTART_CMD);
		system(SMB_RESTART_CMD);
	}
	return 0;
}



/* In main wird die HTML-Ausgabe initialisiert und alles gesteuert */
/*---------------------------------------------------------------------------*/

int main()
{
	struct structExport	*export_list=NULL;	/*Anker der Exportliste */
	char						*env;						/*Parameterstring der vom Web-Server
	                                       	  kommt*/
	int						error=0;					/*Flag das eine Fehlernummer 
															  enthaelt*/
	char						dummy_str[MAXLINE]; 	/*Fuer String-Ergebnisse, die 
															  uninteressant sind*/
	char						help_str[3];
	char						smb_name[MAXLINE];
	char						path[MAXLINE];
	char						type[32];
	int						nfs_root_access=FALSE;
	int						read_only=FALSE;
	char						action[32];
	entry						entries[20];
	int						cl;
	int						x,m,y,i;
	int						parameters,test;
	char						dummy_chr;



	/*debugging*/
	if (DEBUG) header(NONE,"Debug");
   
   /*debugging*/	
   if (DEBUG) printf("main: Env-Variable lesen<br>\n");
   
	/*Die Benutzereingaben einlesen*/
   env=getenv("CONTENT_LENGTH");
   cl = env?atoi(env):0;
	if (cl == 0) error=NO_INPUT_STRING;
	
	/*debugging*/
	if (DEBUG) printf("main: cl: %d, Error: %d<br>",cl,error);
	
	/*Hier werden die Daten eingelesen und in einem Feld des typs "entry" abgelegt*/

	if (!error){
    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
		  if (DEBUG) printf("%d %s %s<br>",x,entries[x].name,entries[x].val);
    }
   }
   parameters=x;
   
   /*debugging*/
	if (DEBUG) printf("main: parameters: %d<br>",parameters);
	   
   /*Die Exporte einlesen*/
 	if (!error) error=read_exports(&export_list);
 
	/*Die einzelnen Parameter herausnehmen und ueberpruefen*/
   /*action wird als erstes ausgewertet, da bei action=aktualisieren
     keine Defaultwerte ermittelt werden duerfen*/
   if (!error) {  
   	if (strcmp(entries[parameters-1].val,"")) strcpy(action,entries[parameters-1].val);
   	else error=NO_ACTION_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d Action: %s<br>",error,action);


	/*Der SMB-Name des Exports*/
   if (!error) {
		strcpy(dummy_str,entries[0].val);
		backtoforward(dummy_str);
		if (containsSlashes(dummy_str)) error=NO_SLASHES_ALLOWED;
		else strcpy(smb_name,dummy_str);
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d smb_name: %s<br>",error,smb_name);

	if (!error) {
	   strcpy(path,"");
	   /*Der Pfad wird ausgewertet*/
		if (strlen(entries[1].val) > 0) {
			if (DEBUG) printf("main: eingegebener Pfad: %s<br>\n",entries[1].val);
			strcpy(dummy_str,entries[1].val);
			backtoforward(dummy_str);
			replaceChr(dummy_str,':','/');
			i=strlen(dummy_str)-1;
			if (dummy_str[i] == '/') dummy_str[i] = '\0';
			if (dummy_str[0] == '/') strcpy(path,dummy_str);
			else {
				strcpy(path,"/");
				strcat(path,dummy_str);
			}
			if (DEBUG) printf("main: ausgewerteter Pfad: %s<br>\n",path);
			if (!DirExists(path) && strcmp(action,"trennen") ) error=PATH_UNKNOWN;
		} else error=PATH_UNKNOWN;
		
      /*debugging*/
		if (DEBUG) printf("main: Error: %d path: %s<br>",error,path);
		
	   /*Der Typ wird ausgewertet*/
		strcpy(type,entries[parameters-2].val);
      /*debugging*/
		if (DEBUG) printf("main: Error: %d type: %s<br>",error,type);

		for (i=2; i<=3; i++) {
			if (!strcmp(entries[i].name,"nfs_root_access")) nfs_root_access=TRUE;
			if (!strcmp(entries[i].name,"read_only")) read_only=TRUE;
			if (DEBUG) printf("main: Error: %d entry-name: %s, read_only: %d, nfs_root_access: %d<br>",error,entries[i].name,read_only,nfs_root_access);
		}

	}
	
	/*debugging*/
	if (DEBUG) printf("Vor der Aktion: Error: %d<br>",error);
	
   		
   /*Hier wird ausgewaehlt, was der User will*/
	
   if (!error){
		if (!strcmp(action,"freigeben")) { 
			if (DEBUG) printf("main: add_export aufrufen<br>");
			error=add_export(export_list, 
			                  smb_name, path, type, read_only, nfs_root_access);
		}
		else { 
			if (!strcmp(action,"trennen") && !error) {
				if (DEBUG) printf("main: delete_export aufrufen<br>");
					error=delete_export(export_list, 
			                  smb_name, path, type, read_only, nfs_root_access);
			}
			else { 
				if (DEBUG) printf("main: change_export aufrufen<br>");
				if (!strcmp(action,"aktualisieren") && !error) 	
					error=change_export(export_list, 
			                  smb_name, path, type, read_only, nfs_root_access);

			}
		}				
   }
   

	/*debugging*/
	printf("Nach der Aktion: Error: %d<br>",error);
	if (DEBUG) tail();

	if (!DEBUG) {
		if (error == NO_ERROR) {
			header(NONE,"Erfolgreich!");
			printf("Die Aktion wurde erfolgreich abgeschlossen\n");
			tail();
		}
		else
			{
				switch (error) {
					case QUERY_STRING_NOT_SET : print_error("Es wurden keine Parameter empfangen ($QUERY_STRING nicht gesetzt).");
					     break;
					case USERID_NOT_NUMBER : print_error("Die Benutzernummer ist keine Zahl");
					     break;
					case GROUP_NOT_NUMBER : print_error("Die Gruppennummer ist keine Zahl");
					     break;
					case USER_EXISTS : print_error("Der Benutzer existiert bereits");
				   	  break;
					case USER_UNKNOWN : print_error("Der Benutzer ist unbekannt");
					     break;
					case USERID_EXISTS : print_error("Die Benutzernummer ist bereits vergeben");
					     break;
					case USERID_UNKNOWN : print_error("Die Benutzernummer ist unbekannt");
				   	  break;
					case GROUP_EXISTS : print_error("Die Gruppe existiert bereits");
					     break;
					case GROUP_UNKNOWN : print_error("Die Gruppe ist unbekannt");
					     break;
					case PASSWD_FAILURE : print_error("Die Datei /etc/passwd konnte nicht gelesen/geschrieben werden");
				   	  break;
					case GROUP_FAILURE : print_error("Die Datei /etc/group konnte nicht gelesen/geschrieben werden");
					     break;
					case NO_GROUPNAME_GIVEN : print_error("Sie m&uuml;ssen einen Gruppennamen angeben!");
					     break;
					case NO_ACTION_GIVEN : print_error("Sie m&uuml;ssen hinzuf&uuml;gen, l&ouml;schen oder &auml;ndern ausw&auml;hlen!");
				   	  break;
					case NO_INPUT_STRING : print_error("Es wurden keine Eingaben aus dem Browser empfangen!");
					     break;
					case PASSWORDS_DIFFERENT : print_error("Die eingegebenen Passw&ouml;rter stimmen nicht &uuml;berein.");
					     break;
					case GROUPID_EXISTS : print_error("Diese Gruppennummer existiert bereits.");
					     break;
					case GROUPID_UNKNOWN : print_error("Diese Gruppennummer ist unbekannt.");
					     break;
					case GROUP_IS_HOME : print_error("Die Gruppe ist die Heimatgruppe eines Benutzers, bitte &auml;ndern sie dessen Heimatgruppe bevor sie die Gruppe l&ouml;schen.");
					     break;
					case FILESYS_FAILURE : print_error("Die Filesystemtabelle konnte nicht geladen werden");
						  break;
					case IMPORT_FAILURE : print_error("Die Import-Tabelle konnte nicht geladen werden");
						  break;
					case NO_DRIVE_GIVEN : print_error("Sie haben vergessen einen Partitionsbuchstaben anzugeben");
						  break;
					case NO_VALID_MOUNTPOINT : print_error("Sie haben ein fehlerhaftes Ziel angegeben (ev. ist das Ziel schon vergeben)");
						  break;
					case DRIVE_ALREADY_MOUNTED : print_error("Dieses Laufwerk ist schon angemeldet");
						  break;
					case DRIVE_UNKNOWN : print_error("Dieses Laufwerk ist im System unbekannt");
						  break;
					case IMPORT_ALREADY_MOUNTED : print_error("Dieser Importname ist bereits vergeben und die betreffende Verbindung ist aktiv");
						  break;
					case IMPORT_UNKNOWN : print_error("Dieser Importname ist im System unbekannt");
						  break;
					case MOUNT_FAILED : print_error("Die Verbindung konnte nicht erstellt werden, &uuml;berpr&uuml;fen sie bitte die Parameter und &auml;ndern sie"
					                                " diese gegebenenfalls.");
						  break;
					case NO_SLASHES_ALLOWED : print_error("Der Name des Importes darf keine \" / \" oder \" \\ \" enthalten");
						  break;
					case PATH_UNKNOWN : print_error("Der Exportpfad konnte nicht gefunden werden.");
						  break;
					case SMB_NAME_EXISTS : print_error("Dieser Windows-Name ist bereits vergeben.");
						  break;
					case EXPORT_EXISTS : print_error("Dieser Pfad ist bereits exportiert.");
						  break;
					case EXPORT_UNKNOWN : print_error("Dieser Pfad ist nicht exportiert.");
						  break;
					case SMB_NAME_NOT_SET : print_error("Wenn ein Verzeichnis Windows-kompatibel exportiert werden soll, muss ein Windows-Name angegeben werden.");
						  break;
				}
			}
	}
	return 0;
}
