/******************************************************************************
* Dieses Programm dient zum aendern der Benutzerdaten (/etc/passwd) auf einem *
* Unix-Rechner																						*
******************************************************************************/

/* Die Includes werden gemacht */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <support.h>
#include <string.h>
#include "util.h"



/*Lokale Defines */
/*---------------------------------------------------------------------------*/

/*in print-error wird eine Fehlerseite generiert und ausgegeben*/
/*---------------------------------------------------------------------------*/
void print_error(char* error){

	header(NONE,"Fehler");
	printf("%s\n",error);
	tail();
}


/*in add_user wird ein Benutzer eingefuegt*/
/*---------------------------------------------------------------------------*/
int add_group(struct structUser *user_base, 
             struct structGroup *group_base,
             char groupname[MAXLINE],
             long int groupid,
	          char add_members[MAXLINE],
	          char delete_members[MAXLINE]) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE					*passwd_file;			/*File-Handle um in /etc/passwd zu
	                                         schreiben*/
	FILE					*group_file;			/*File-Handle um in /etc/group zu
														  schreiben*/
	struct structGroup *prev;
	struct structGroup *next;
	struct structGroup *akt;
	char					*buffer;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   
   /*Die Parameter auf Korrektheit pruefen*/
	if (GidExists(group_base, groupid)) return GROUPID_EXISTS; 
	if (GnameExists(group_base,groupname)) return GROUP_EXISTS; 	

	/*Die neue Gruppe wird in die Gruppenliste eingefuegt*/
	new_group(group_base,groupid,groupname);

	akt=group_base;
	while (akt->gid != groupid) akt=akt->next;

	buffer=(char*) malloc(sizeof(char)*MAXLINE);
	while (strlen(add_members) > 0) {
		getword(buffer,add_members,',');
		if (UserExists(user_base,buffer)) addUserToGroup(akt,buffer);
	}
	free(buffer);
	
	/*/etc/group wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/group /etc/group.old");

	if (!(group_file=fopen(GROUP_NAME,"wb"))) {
		return GROUP_FAILURE;
	}

	/*Die Gruppenliste wird geschrieben*/
	akt=group_base;
	do{
		fprintf(group_file,"%s:*:%d:%s\n",akt->groupname,akt->gid,akt->users);
		akt=akt->next;
	}while (akt != group_base);
	fclose(group_file);
}


/*in delete_group wird eine Gruppe geloescht*/
/*---------------------------------------------------------------------------*/
int delete_group(struct structUser *user_base, 
             struct structGroup *group_base,
             char groupname[MAXLINE],
             long int groupid,
	          char add_members[MAXLINE],
	          char delete_members[MAXLINE]) {
	          
	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE						*passwd_file;	/*Filehandle fuer /etc/passwd*/
	FILE						*group_file;	/*Filehandle fuer /etc/group*/
	struct structGroup 	*prev;			/*Pointer zum Manipulieren*/
	struct structGroup 	*next;         /*der Gruppenliste*/
	struct structGroup	*akt;
  	struct structUser		*akt_user;
   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	/*Wenn die Gruppe nicht existiert kann sie nicht geloescht werden*/
	if (!GnameExists(group_base, groupname)) return GROUP_UNKNOWN;
	
	/*Es wird nach der Gruppe gesucht*/
	akt=group_base;
	while(strcmp(akt->groupname,groupname)) akt=akt->next;
	
	akt_user=user_base;
	do {
		if (akt_user->gid == akt->gid) return GROUP_IS_HOME;
		akt_user=akt_user->next;
	} while (akt_user != user_base);  
	
	/*Die Gruppe wird aus der Gruppenliste entfernt*/
	
	prev=akt->prev;
	next=akt->next;
	prev->next=next;
	next->prev=prev;

	/*Freigeben des Speichers*/
	if (group_base->next == group_base) group_base=NULL;
	free(akt);

	if (group_base != NULL) {
		/*/etc/group wird gesichert und zum schreiben geoeffnet*/
		system("mv /etc/group /etc/group.old");
		
		if (!(group_file=fopen(GROUP_NAME,"wb"))) {
			return GROUP_FAILURE;
		}
	
		/*Die Gruppenliste wird geschrieben*/
		akt=group_base;
		do{
			fprintf(group_file,"%s:*:%d:%s\n",akt->groupname,akt->gid,akt->users);
			akt=akt->next;
		}while (akt != group_base);
		fclose(group_file);
	}
}


/*mit change_group koennen Gruppendaten veraendert werden*/
/*---------------------------------------------------------------------------*/
int change_group(struct structUser *user_base, 
             struct structGroup *group_base,
             char groupname[MAXLINE],
             long int groupid,
	          char add_members[MAXLINE],
	          char delete_members[MAXLINE]) {
	          
	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE						*passwd_file;	/*Filehandle fuer /etc/passwd*/
	FILE						*group_file;	/*Filehandle fuer /etc/group*/
	struct structGroup 	*prev;			/*Pointer zum Manipulieren*/
	struct structGroup 	*next;         /*der Gruppenliste*/
	struct structGroup	*akt;
  	struct structUser		*akt_user;
	char						*buffer;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   
	/*debugging*/
	if (DEBUG) printf("change_group: Vor der Parameterpruefung<br>");
   
   /*Die Parameter auf Korrektheit pruefen*/
	if (!GnameExists(group_base, groupname)) return GROUP_UNKNOWN;
	if ((groupid != -1) && (GidExists(group_base, groupid))) return GROUPID_EXISTS; 
	
	/*debugging*/
	if (DEBUG) printf("Nach der Parameterpruefung<br>");
		
	/*Die Gruppe wird gesucht*/
	akt=group_base;
	while (strcmp(akt->groupname,groupname)) akt=akt->next;
	
	/*debugging*/
	if (DEBUG) printf("change_group: Gruppe gefunden<br>");

	/*Neue Gid setzen*/
	if (groupid != -1) akt->gid=groupid;

	/*Benutzer entfernen und hinzufuegen*/
	buffer=(char*) malloc(sizeof(char)*MAXLINE);
	while (strlen(delete_members) > 0) {
		getword(buffer,delete_members,',');
		if (UserExists(user_base,buffer)) delUserFromGroup(akt,buffer);
	}
	
	while (strlen(add_members) > 0) {
		getword(buffer,add_members,',');
		if (UserExists(user_base,buffer)) addUserToGroup(akt,buffer);
	}
	free(buffer);

	
	/*/etc/group wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/group /etc/group.old");

	if (!(group_file=fopen(GROUP_NAME,"wb"))) {
		return GROUP_FAILURE;
	}

	/*Die Gruppenliste wird geschrieben*/
	akt=group_base;
	do{
		fprintf(group_file,"%s:*:%d:%s\n",akt->groupname,akt->gid,akt->users);
		akt=akt->next;
	}while (akt != group_base);
	fclose(group_file);
}



/* In main wird die HTML-Ausgabe initialisiert und alles gesteuert */
/*---------------------------------------------------------------------------*/

int main()
{
	struct structUser 	*user_list=NULL;		/*Anker der Userliste */
	struct structGroup 	*group_list=NULL;		/*Anker der Gruppenliste */
	char						*env;						/*Parameterstring der vom Web-Server
	                                       	  kommt*/
	int						error=0;					/*Flag das eine Fehlernummer 
															  enthaelt*/
	char						dummy_str[MAXLINE]; 	/*Fuer String-Ergebnisse, die 
															  uninteressant sind*/
	char						groupname[MAXLINE];	/*Die einzelnen Parameter*/
	char						group_str[MAXLINE];
	long int						group;
	char						add_members[MAXLINE];
	char						delete_members[MAXLINE];
	char						action[32];
	entry						entries[MAXENTRIES];
	int						cl;
	int						x,m;
	int						parameters;


	/*debugging*/
	if (DEBUG) header(GRUPPEN,"Debug");
   
   /*debugging*/	
   if (DEBUG) printf("main: Env-Variable lesen<br>\n");
   
	/*Die Benutzereingaben einlesen*/
   env=getenv("CONTENT_LENGTH");
   cl = env?atoi(env):0;
	if (cl == 0) error=NO_INPUT_STRING;
	
	/*debugging*/
	if (DEBUG) printf("main: cl: %d, Error: %d<br>",cl,error);
	
	/*Hier werden die Daten eingelesen und in einem Feld des typs "entry" abgelegt*/

	if (!error){
    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
    }
   }
   parameters=x;
   
   /*debugging*/
	if (DEBUG) printf("main: parameters: %d<br>",parameters);
	   
	/*Die passwd-Datei einlesen*/
	if (!error) error=read_users(&user_list);
   
      
   /*Die group-Datei einlesen*/
 	if (!error) error=read_groups(&group_list);

	/*Die einzelnen Parameter herausnehmen und ueberpruefen*/
   /*action wird als erstes ausgewertet, da bei action=aktualisieren
     keine Defaultwerte ermittelt werden duerfen*/
   if (!error) {  
   	if (strcmp(entries[4].val,"")) strcpy(action,entries[4].val);
   	else error=NO_ACTION_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("Action: %s<br>",action);

	if (!DEBUG) {

	/*Der Groupname wird ausgewertet*/
   if (!error) {
		strcpy(groupname,entries[0].val);
		if (!strcmp(groupname,"")) error=NO_GROUPNAME_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("groupname: %s<br>",groupname);

	if (!error) {
	   
	   /*Die GroupId wird ausgewertet*/
		strcpy(group_str,entries[1].val);
		if (strcmp(group_str,"")){
			if (!isNumber(group_str) && !error) error=GROUP_NOT_NUMBER;
			if (!error) group=atoi(group_str);
		}
		else {
			if (strcmp(action,"aktualisieren")) group=get_unused_gid(group_list);
         else group=-1;
      }
      /*debugging*/
		if (DEBUG) printf("main: group: %d<br>",group);

		/*Die einzufuegenden Mitglieder werden eingelesen*/
		strcpy(add_members,entries[2].val);

		/*debugging*/
		if (DEBUG) printf("main: add_members: %s<br>",add_members);

		/*Die zu entfernenden Mitglieder werden eingelesen*/
		strcpy(delete_members,entries[3].val);

		/*debugging*/
		if (DEBUG) printf("main: delete_members: %s<br>",delete_members);
	}	
	
	}
	/*debugging*/
	if (DEBUG) printf("Vor der Aktion: Error: %d<br>",error);
	
		
   /*Hier wird ausgewaehlt, was der User will*/
   if (!error){
		if (!strcmp(action,"eintragen")) { 
			if (DEBUG) printf("main: add_user aufrufen<br>");
			error=add_group(user_list, group_list, groupname, group, add_members, delete_members);
		}
		else { 
			if (!strcmp(action,"entfernen") && !error) {
				if (DEBUG) printf("main: delete_user aufrufen<br>");
				error=delete_group(user_list, group_list, groupname, group, add_members, delete_members);
			}
			else { 
				if (DEBUG) printf("main: change_user aufrufen<br>");
				if (!strcmp(action,"aktualisieren") && !error) error=change_group(user_list, group_list, groupname, group, add_members, delete_members);
			}
		}				
   }
   

	/*debugging*/
	printf("Nach der Aktion: Error: %d<br>",error);
	if (DEBUG) tail();

	if (!DEBUG) {
		if (error == NO_ERROR) {
			header(NONE,"Erfolgreich!");
			printf("Die Aktion wurde erfolgreich abgeschlossen\n");
			tail();
		}
		else
			{
				switch (error) {
					case QUERY_STRING_NOT_SET : print_error("Es wurden keine Parameter empfangen ($QUERY_STRING nicht gesetzt).");
					     break;
					case USERID_NOT_NUMBER : print_error("Die Benutzernummer ist keine Zahl");
					     break;
					case GROUP_NOT_NUMBER : print_error("Die Gruppennummer ist keine Zahl");
					     break;
					case USER_EXISTS : print_error("Der Benutzer existiert bereits");
				   	  break;
					case USER_UNKNOWN : print_error("Der Benutzer ist unbekannt");
					     break;
					case USERID_EXISTS : print_error("Die Benutzernummer ist bereits vergeben");
					     break;
					case USERID_UNKNOWN : print_error("Die Benutzernummer ist unbekannt");
				   	  break;
					case GROUP_EXISTS : print_error("Die Gruppe existiert bereits");
					     break;
					case GROUP_UNKNOWN : print_error("Die Gruppe ist unbekannt");
					     break;
					case PASSWD_FAILURE : print_error("Die Datei /etc/passwd konnte nicht gelesen/geschrieben werden");
				   	  break;
					case GROUP_FAILURE : print_error("Die Datei /etc/group konnte nicht gelesen/geschrieben werden");
					     break;
					case NO_GROUPNAME_GIVEN : print_error("Sie m&uuml;ssen einen Gruppennamen angeben!");
					     break;
					case NO_ACTION_GIVEN : print_error("Sie m&uuml;ssen hinzuf&uuml;gen, l&ouml;schen oder &auml;ndern ausw&auml;hlen!");
				   	  break;
					case NO_INPUT_STRING : print_error("Es wurden keine Eingaben aus dem Browser empfangen!");
					     break;
					case PASSWORDS_DIFFERENT : print_error("Die eingegebenen Passw&ouml;rter stimmen nicht &uuml;berein.");
					     break;
					case GROUPID_EXISTS : print_error("Diese Gruppennummer existiert bereits.");
					     break;
					case GROUPID_UNKNOWN : print_error("Diese Gruppennummer ist unbekannt.");
					     break;
					case GROUP_IS_HOME : print_error("Die Gruppe ist die Heimatgruppe eines Benutzers, bitte &auml;ndern sie dessen Heimatgruppe bevor sie die Gruppe l&ouml;schen.");
					     break;
				}
			}
	}
	return 0;
}
