/******************************************************************************
* Aendern von /etc/fstab																	   *
* 																										*
******************************************************************************/

/* Die Includes werden gemacht */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <support.h>
#include <string.h>
#include <ctype.h>
#include "util.h"



/*Lokale Defines */
/*---------------------------------------------------------------------------*/

/*in print-error wird eine Fehlerseite generiert und ausgegeben*/
/*---------------------------------------------------------------------------*/
void print_error(char* error){

	header(NONE,"Fehler");
	printf("%s\n",error);
	tail();
}


/*in add_import wird ein Import in fstab eingetragen*/
/*---------------------------------------------------------------------------*/
int add_import(struct structUser *user_base, 
             struct structGroup *group_base,
             struct structFilesys *filesys_base,
             struct structImport *import_base,
             char server[MAXLINE],
             char	serverpath[MAXLINE],
             char	mountpoint[MAXLINE],
             char type[10],
             int	blocksize,
             int	retries,
             int	bootloaded) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structImport *akt;
	char					*cmdstring;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   
	if (DEBUG) printf("add_import: Import: %s bootloaded %d<br>",mountpoint,bootloaded);
	/*Wenn der Import existiert und aktiv ist liegt ein Fehler vor*/
	if (ImportExists(import_base,mountpoint)) {
			if (DEBUG) printf("add_import: Import ist bereits vorhanden<br>");
			if (ImportMounted(import_base,mountpoint)) {
				if (DEBUG) printf("add_import: Import ist bereits aktiv<br>");
				return IMPORT_ALREADY_MOUNTED;
			}
	}


	/*Wenn der Import bereits existiert, aber nicht aktiv ist, wird versucht die
	  Verbindung zu erstellen, anderenfalls wird der Import in die Importliste
	  eingetragen und ein Verbindungsversuch gestartet*/	
   
	if (ImportExists(import_base,mountpoint)) {
		
		akt = GetImport(import_base , mountpoint);
		sprintf(cmdstring,"mount_nfs -o wsize=%d -o rsize=%d -o retrans=%d %s:%s %s",
		        akt->blocksize,akt->blocksize,akt->retries,akt->server,akt->serverpath,
	   	     akt->mountpoint);
		system(cmdstring);
		if (!DirExists(akt->mountpoint)) return MOUNT_FAILED; 
		
	} else {
	   if (DEBUG) printf("add_import: Import wird neu angelegt: %s<br>",mountpoint);
	   akt = (struct structImport *) malloc(sizeof(struct structImport));
		strcpy(akt->server,server);
		strcpy(akt->serverpath,serverpath);
		strcpy(akt->mountpoint,mountpoint);
		strcpy(akt->type,type);
		akt->blocksize=blocksize;
		akt->retries=retries;
		akt->bootloaded=bootloaded;
		akt->mounted=FALSE;
		addImport(&import_base,akt);
		
		FlushFstab(filesys_base,import_base);

		cmdstring=(char *) malloc(sizeof(char)*MAXLINE);
		sprintf(cmdstring,"mount_nfs -o wsize=%d -o rsize=%d -o retrans=%d %s:%s %s",
		        akt->blocksize,akt->blocksize,akt->retries,akt->server,akt->serverpath,
		        akt->mountpoint);
		system(cmdstring);
		if (!DirExists(akt->mountpoint)) return MOUNT_FAILED; 
		
	}
	
	return 0;
}


/*in delete_import wird ein Import aus der fstab geloescht*/
/*---------------------------------------------------------------------------*/
int delete_import(struct structUser *user_base, 
             struct structGroup *group_base,
             struct structFilesys *filesys_base,
             struct structImport *import_base,
             char server[MAXLINE],
             char	serverpath[MAXLINE],
             char	mountpoint[MAXLINE],
             char type[10],
             int	blocksize,
             int	retries,
             int	bootloaded) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (!ImportExists(import_base,mountpoint)) return IMPORT_UNKNOWN;
	if (DEBUG) printf("delete_import: Aufruf von delImport()<br>");
	delImport(&import_base,mountpoint);
	if (DEBUG) printf("delete_import: Aufruf von FlushFstab()<br>");
	FlushFstab(filesys_base,import_base);
}


/*mit change_import koennen die mount-Parameter veraendert werden*/
/*---------------------------------------------------------------------------*/
int change_import(struct structUser *user_base, 
             struct structGroup *group_base,
             struct structFilesys *filesys_base,
             struct structImport *import_base,
             char server[MAXLINE],
             char	serverpath[MAXLINE],
             char	mountpoint[MAXLINE],
             char type[10],
             int	blocksize,
             int	retries,
             int	bootloaded) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structImport *akt;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (!ImportExists(import_base,mountpoint)) return IMPORT_UNKNOWN;
	akt = GetImport(import_base,mountpoint);
	if (DEBUG) printf("change_import: mountpoint: %s, type: %s, server: %s, serverpath: %s<br>",
							mountpoint,type,server,serverpath);
	if (DEBUG) printf("change_import: akt->mountpoint: %s, akt->type: %s, akt->server: %s, akt->serverpath: %s<br>",
							akt->mountpoint,akt->type,akt->server,akt->serverpath);
	if (strcmp(server,"")) strcpy(akt->server,server);
	if (strcmp(serverpath,"/")) strcpy(akt->serverpath,serverpath);
	if (strcmp(type,"")) strcpy(akt->type,type);
	if (blocksize != 0) akt->blocksize=blocksize;
	if (retries != 0) akt->retries=retries;
	if (bootloaded) akt->bootloaded = bootloaded;
	FlushFstab(filesys_base,import_base);
}



/* In main wird die HTML-Ausgabe initialisiert und alles gesteuert */
/*---------------------------------------------------------------------------*/

int main()
{
	struct structUser 	*user_list=NULL;		/*Anker der Userliste */
	struct structGroup 	*group_list=NULL;		/*Anker der Gruppenliste */
	struct structFilesys	*filesys_list=NULL;	/*Anker der Filesystemliste */
	struct structImport	*import_list=NULL;	/*Anker der Importliste */
	char						*env;						/*Parameterstring der vom Web-Server
	                                       	  kommt*/
	int						error=0;					/*Flag das eine Fehlernummer 
															  enthaelt*/
	char						dummy_str[MAXLINE]; 	/*Fuer String-Ergebnisse, die 
															  uninteressant sind*/
	char						help_str[3];
	char						server[MAXLINE];
	char						serverpath[MAXLINE];
	char						mountpoint[MAXLINE];
	char						type[10];
	int						blocksize;
	int						retries;
	int						bootloaded=FALSE;
	char						action[32];
	entry						entries[20];
	int						cl;
	int						x,m,y;
	int						parameters,test;
	char						dummy_chr;



	/*debugging*/
	if (DEBUG) header(NONE,"Debug");
   
   /*debugging*/	
   if (DEBUG) printf("main: Env-Variable lesen<br>\n");
   
	/*Die Benutzereingaben einlesen*/
   env=getenv("CONTENT_LENGTH");
   cl = env?atoi(env):0;
	if (cl == 0) error=NO_INPUT_STRING;
	
	/*debugging*/
	if (DEBUG) printf("main: cl: %d, Error: %d<br>",cl,error);
	
	/*Hier werden die Daten eingelesen und in einem Feld des typs "entry" abgelegt*/

	if (!error){
    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
		  if (DEBUG) printf("%d %s %s<br>",x,entries[x].name,entries[x].val);
    }
   }
   parameters=x;
   
   /*debugging*/
	if (DEBUG) printf("main: parameters: %d<br>",parameters);
	   
	/*Die passwd-Datei einlesen*/
	if (!error) error=read_users(&user_list);
   
      
   /*Die group-Datei einlesen*/
 	if (!error) error=read_groups(&group_list);

   /*Die filesysteme einlesen*/
 	if (!error) error=read_filesys(&filesys_list);

   /*Die imports einlesen*/
 	if (!error) error=read_imports(&import_list);
 
	/*Die einzelnen Parameter herausnehmen und ueberpruefen*/
   /*action wird als erstes ausgewertet, da bei action=aktualisieren
     keine Defaultwerte ermittelt werden duerfen*/
   if (!error) {  
   	if (strcmp(entries[parameters-1].val,"")) strcpy(action,entries[parameters-1].val);
   	else error=NO_ACTION_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d Action: %s<br>",error,action);


	/*Der Name des Imports*/
   if (!error) {
		strcpy(dummy_str,entries[0].val);
		backtoforward(dummy_str);
		if (containsSlashes(dummy_str)) error=NO_SLASHES_ALLOWED;
		if (!strcmp(dummy_str,"")) error=IMPORT_UNKNOWN;
		else {
			strcpy(mountpoint,"/nfs/");
			strcat(mountpoint,dummy_str);
		}

   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d Mountpoint: %s<br>",error,mountpoint);

	if (!error) {

		if (parameters == 7) bootloaded=TRUE;
		if (DEBUG) printf("main: Error: %d server: %d<br>",error,bootloaded);
		
	   
	   /*Der Servername wird ausgewertet*/
		strcpy(server,entries[parameters-5].val);
      /*debugging*/
		if (DEBUG) printf("main: Error: %d server: %s<br>",error,server);
		
	   /*Der Serverpfad wird ausgewertet*/
		strcpy(dummy_str,entries[parameters-4].val);
		/*eine eventuell notwendige Konvertierung ins UNIX-format wird vorgenommen*/
		x=0;
		backtoforward(dummy_str); 
		if (dummy_str[0] != '/') strcpy(serverpath,"/");
		else strcpy(serverpath,"");
		while (x < strlen(dummy_str)) {
			if (dummy_str[x] != ':') {
				help_str[0] = dummy_str[x];
				help_str[1] = '\0';
				strcat(serverpath,help_str);
				
				if (DEBUG) printf("main: Konvertieren des serverpfades: %s<br>\n",serverpath);
			}
			x++;
		}
      /*debugging*/
		if (DEBUG) printf("main: Error: %d serverpath: %s<br>",error,serverpath);

	   /*Der Typ wird ausgewertet*/
		strcpy(type,"nfs");
      /*debugging*/
		if (DEBUG) printf("main: Error: %d type: %s<br>",error,type);

	   /*Die Blockgroesse wird ausgewertet*/
		strcpy(dummy_str,entries[parameters-3].val);
		if (strcmp(dummy_str,"keine Angabe")) blocksize=atoi(dummy_str);
		else	if (!strcmp(action,"aktualisieren")) blocksize=0;
				else blocksize=DEF_BLOCKSIZE;
      /*debugging*/
		if (DEBUG) printf("main: Error: %d blocksize: %d<br>",error,blocksize);
	
	   /*Die Anzahl der Retries wird ausgewertet*/
		strcpy(dummy_str,entries[parameters-2].val);
		if (strcmp(dummy_str,"keine Angabe")) retries=atoi(dummy_str);
		else	if (!strcmp(action,"aktualisieren")) retries=0;
				else retries=DEF_RETRANS;
      /*debugging*/
		if (DEBUG) printf("main: Error: %d retries: %d<br>",error,retries);

	}
	/*debugging*/
	if (DEBUG) printf("Vor der Aktion: Error: %d<br>",error);
	
   		
   /*Hier wird ausgewaehlt, was der User will*/
	
   if (!error){
		if (!strcmp(action,"verbinden")) { 
			if (DEBUG) printf("main: add_import aufrufen<br>");
			error=add_import(user_list, group_list, filesys_list, import_list, 
			                  server, serverpath, mountpoint, type, blocksize, retries, bootloaded);
		}
		else { 
			if (!strcmp(action,"entfernen") && !error) {
				if (DEBUG) printf("main: delete_import aufrufen<br>");
					error=delete_import(user_list, group_list, filesys_list, import_list, 
			                  server, serverpath, mountpoint, type, blocksize, retries, bootloaded);
			}
			else { 
				if (DEBUG) printf("main: change_import aufrufen<br>");
				if (!strcmp(action,"aktualisieren") && !error) 	
					error=change_import(user_list, group_list, filesys_list, import_list, 
			      							server, serverpath, mountpoint, type, blocksize, retries, bootloaded);

			}
		}				
   }
   

	/*debugging*/
	printf("Nach der Aktion: Error: %d<br>",error);
	if (DEBUG) tail();

	if (!DEBUG) {
		if (error == NO_ERROR) {
			header(NONE,"Erfolgreich!");
			printf("Die Aktion wurde erfolgreich abgeschlossen\n");
			tail();
		}
		else
			{
				switch (error) {
					case QUERY_STRING_NOT_SET : print_error("Es wurden keine Parameter empfangen ($QUERY_STRING nicht gesetzt).");
					     break;
					case USERID_NOT_NUMBER : print_error("Die Benutzernummer ist keine Zahl");
					     break;
					case GROUP_NOT_NUMBER : print_error("Die Gruppennummer ist keine Zahl");
					     break;
					case USER_EXISTS : print_error("Der Benutzer existiert bereits");
				   	  break;
					case USER_UNKNOWN : print_error("Der Benutzer ist unbekannt");
					     break;
					case USERID_EXISTS : print_error("Die Benutzernummer ist bereits vergeben");
					     break;
					case USERID_UNKNOWN : print_error("Die Benutzernummer ist unbekannt");
				   	  break;
					case GROUP_EXISTS : print_error("Die Gruppe existiert bereits");
					     break;
					case GROUP_UNKNOWN : print_error("Die Gruppe ist unbekannt");
					     break;
					case PASSWD_FAILURE : print_error("Die Datei /etc/passwd konnte nicht gelesen/geschrieben werden");
				   	  break;
					case GROUP_FAILURE : print_error("Die Datei /etc/group konnte nicht gelesen/geschrieben werden");
					     break;
					case NO_GROUPNAME_GIVEN : print_error("Sie m&uuml;ssen einen Gruppennamen angeben!");
					     break;
					case NO_ACTION_GIVEN : print_error("Sie m&uuml;ssen hinzuf&uuml;gen, l&ouml;schen oder &auml;ndern ausw&auml;hlen!");
				   	  break;
					case NO_INPUT_STRING : print_error("Es wurden keine Eingaben aus dem Browser empfangen!");
					     break;
					case PASSWORDS_DIFFERENT : print_error("Die eingegebenen Passw&ouml;rter stimmen nicht &uuml;berein.");
					     break;
					case GROUPID_EXISTS : print_error("Diese Gruppennummer existiert bereits.");
					     break;
					case GROUPID_UNKNOWN : print_error("Diese Gruppennummer ist unbekannt.");
					     break;
					case GROUP_IS_HOME : print_error("Die Gruppe ist die Heimatgruppe eines Benutzers, bitte &auml;ndern sie dessen Heimatgruppe bevor sie die Gruppe l&ouml;schen.");
					     break;
					case FILESYS_FAILURE : print_error("Die Filesystemtabelle konnte nicht geladen werden");
						  break;
					case IMPORT_FAILURE : print_error("Die Import-Tabelle konnte nicht geladen werden");
						  break;
					case NO_DRIVE_GIVEN : print_error("Sie haben vergessen einen Partitionsbuchstaben anzugeben");
						  break;
					case NO_VALID_MOUNTPOINT : print_error("Sie haben ein fehlerhaftes Ziel angegeben (ev. ist das Ziel schon vergeben)");
						  break;
					case DRIVE_ALREADY_MOUNTED : print_error("Dieses Laufwerk ist schon angemeldet");
						  break;
					case DRIVE_UNKNOWN : print_error("Dieses Laufwerk ist im System unbekannt");
						  break;
					case IMPORT_ALREADY_MOUNTED : print_error("Dieser Importname ist bereits vergeben und die betreffende Verbindung ist aktiv");
						  break;
					case IMPORT_UNKNOWN : print_error("Dieser Importname ist im System unbekannt");
						  break;
					case MOUNT_FAILED : print_error("Die Verbindung konnte nicht erstellt werden, &uuml;berpr&uuml;fen sie bitte die Parameter und &auml;ndern sie"
					                                " diese gegebenenfalls.");
						  break;
					case NO_SLASHES_ALLOWED : print_error("Der Name des Importes darf keine \" / \" oder \" \\ \" enthalten");
						  break;
				}
			}
	}
	return 0;
}
