/******************************************************************************
* Netzwerk-Exporte bearbeiten																   *
* 																										*
******************************************************************************/

/* Die Includes werden gemacht */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <support.h>
#include <string.h>
#include <ctype.h>
#include "util.h"



/*Lokale Defines */
/*---------------------------------------------------------------------------*/

/*in print-error wird eine Fehlerseite generiert und ausgegeben*/
/*---------------------------------------------------------------------------*/
void print_error(char* error){

	header(NONE,"Fehler");
	printf("%s\n",error);
	tail();
}



/*---------------------------------------------------------------------------*/
int add_printer(struct structPrinter *printer_base,
					char	name[MAXLINE],
             	char	type[MAXLINE],
             	char	printserver[MAXLINE],
             	char	remotename[MAXLINE],
             	int	formfeed,
             	int	headerpage,
             	int	max,
             	char	rpasswd[MAXLINE]) {
					

	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structPrinter *akt;
	char					*cmdstring;
	char					smbconf_path[MAXLINE];
	FILE					*smbconf_file;
	char					spooldir[MAXLINE];
	int					spool_len;
	char					logfilename[MAXLINE];

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (PrinterExists(printer_base,name)) return PRINTER_ALREADY_EXISTS;
	akt = (struct structPrinter *) malloc(sizeof(struct structPrinter));
	cmdstring = (char *) malloc(sizeof(char)*MAXLINE);
	
	/*Bei einem neu eingerichteten Drucker gibt es keine unknown options*/
	strcpy(akt->unknown_options,"");
	 
	strcpy(akt->name,name);
	strcpy(akt->sd,PRINTSPOOL_ROOT);
	strcat(akt->sd,name);
	strcat(akt->sd,"/");

	/*Das Spoolverzeichnis erstellen*/
	strcpy(spooldir,akt->sd);
	spool_len=strlen(spooldir);
	spooldir[spool_len-1]='\0';
	
	/*Einrichten des Directorys*/
	sprintf(cmdstring,"mkdir %s",spooldir);
	if (DEBUG) printf("add_printer: cmdstring: %s<br>\n",cmdstring);
	system(cmdstring);
	sprintf(cmdstring,"chmod 777 %s",spooldir);
	if (DEBUG) printf("add_printer: cmdstring: %s<br>\n",cmdstring);
	system(cmdstring);
	sprintf(cmdstring,"chgrp wheel %s",spooldir);
	if (DEBUG) printf("add_printer: cmdstring: %s<br>\n",cmdstring);
	system(cmdstring);
	strcpy(logfilename,akt->sd);
	strcat(logfilename,"printer-log");
	sprintf(cmdstring,"touch %s",logfilename);
	if (DEBUG) printf("add_printer: cmdstring: %s<br>\n",cmdstring);
	system(cmdstring);


	if (!strcmp(type,"Netz (smb)")) {
		if (!strlen(printserver)) return PRINTSERVER_MISSING;
		if (!strlen(remotename)) return REMOTENAME_MISSING;
		strcpy(akt->infilter,SMBPRINT_NAME);
		strcpy(akt->type,"smb");
		strcpy(smbconf_path,akt->sd);
		strcat(smbconf_path,".config");
		if (!(smbconf_file=fopen(smbconf_path,"wb"))) {
			return SMBCONF_FAILURE;
		} 
		fprintf(smbconf_file,"server=%s\nservice=%s\npassword=\"%s\"\n",printserver,remotename,rpasswd);
		fclose(smbconf_file);
	} else strcpy(akt->infilter,"");

	if (!strcmp(type,"Netz (bsd)")) {
		if (!strlen(printserver)) return PRINTSERVER_MISSING;
		strcpy(akt->type,"bsd");
		strcpy(akt->rm,printserver);
		strcpy(akt->rp,remotename);
		strcpy(akt->lp,"");
	} else {
		strcpy(akt->rm,"");
		strcpy(akt->rp,"");
	}
	
		
	if (strcmp(type,"Netz (smb)") && strcmp(type,"Netz (nfs)")) {
		strcpy(akt->lp,"/dev/lp");
		strcpy(akt->type,"local");
	}
	else strcpy(akt->lp,"/dev/null");
	akt->sf=formfeed;
	akt->sh=headerpage;
	akt->mx=max;
	addPrinter(&printer_base,akt);
	FlushPrinters(printer_base);
	system(LPSPOOL_RESTART_CMD);
	return 0;
}



/*---------------------------------------------------------------------------*/
int delete_printer(struct structPrinter *printer_base,
					char	name[MAXLINE],
             	char	type[MAXLINE],
             	char	printserver[MAXLINE],
             	char	remotename[MAXLINE],
             	int	formfeed,
             	int	headerpage,
             	int	max,
             	char	rpasswd[MAXLINE]) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structPrinter *akt;
	char					*cmdstring;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (!PrinterExists(printer_base,name)) return PRINTER_UNKNOWN;
	akt = GetPrinter(printer_base,name);
	cmdstring = (char *) malloc(sizeof(char)*MAXLINE);
	if (strlen(akt->sd)) {
		sprintf(cmdstring,"rm -r %s",akt->sd);
		system(cmdstring);
	}
	delPrinter(&printer_base,name);
	FlushPrinters(printer_base);   
	system(LPSPOOL_RESTART_CMD);
	return 0;
}



/*---------------------------------------------------------------------------*/
int change_printer(struct structPrinter *printer_base,
					char	name[MAXLINE],
             	char	type[MAXLINE],
             	char	printserver[MAXLINE],
             	char	remotename[MAXLINE],
             	int	formfeed,
             	int	headerpage,
             	int	max,
             	char	rpasswd[MAXLINE]) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	struct structPrinter *akt;
	char					*cmdstring;
	char					smbconf_path[MAXLINE];
	FILE					*smbconf_file;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (!PrinterExists(printer_base,name)) return PRINTER_UNKNOWN;
	akt = GetPrinter(printer_base,name);


	if (!strcmp(type,"Lokal")) strcpy(akt->type,"local"); 
	if (strlen(printserver)) strcpy(akt->rm,printserver);
	if (strlen(remotename)) strcpy(akt->rp,remotename);
	if (formfeed != -1) akt->sf=formfeed;
	if (headerpage != -1) akt->sh=headerpage;
	if (max != -1) akt->mx=max;
	if (!strcmp(type,"Netz (smb)") 
	 || (!strcmp(akt->type,"smb") 
	  && (strlen(printserver) || strlen(remotename) || strlen(rpasswd))
	    )
	   ) {
		strcpy(akt->type,"smb");
		strcpy(smbconf_path,akt->sd);
		strcat(smbconf_path,".config");
		if (!(smbconf_file=fopen(smbconf_path,"wb"))) {
			return SMBCONF_FAILURE;
		} 
		fprintf(smbconf_file,"server=%s\nservice=%s\npassword=%s\n",akt->rm,akt->rp,rpasswd);
		fclose(smbconf_file);
		} 
	if (!strcmp(type,"Netz (bsd)")) {
		strcpy(akt->type,"bsd");
		if (!strlen(printserver)) return PRINTSERVER_MISSING;
		if (!strlen(remotename)) strcpy(akt->rp,"");
		strcpy(akt->lp,"");
	}	
	FlushPrinters(printer_base);	
	system(LPSPOOL_RESTART_CMD);
	return 0;
}



/* In main wird die HTML-Ausgabe initialisiert und alles gesteuert */
/*---------------------------------------------------------------------------*/

int main()
{
	struct structPrinter	*printer_list=NULL;	/*Anker der Exportliste */
	char						*env;						/*Parameterstring der vom Web-Server
	                                       	  kommt*/
	int						error=0;					/*Flag das eine Fehlernummer 
															  enthaelt*/
	char						dummy_str[MAXLINE]; 	/*Fuer String-Ergebnisse, die 
															  uninteressant sind*/
	char						help_str[3];
	char						name[MAXLINE];
	char						type[MAXLINE];
	char						printserver[MAXLINE];
	char						remotename[MAXLINE];
	char						rpasswd[MAXLINE];
	int						max;
	int						formfeed=TRUE;
	int						headerpage=TRUE;
	char						action[32];
	entry						entries[20];
	int						cl;
	int						x,m,y,i;
	int						parameters,test;
	char						dummy_chr;



	/*debugging*/
	if (DEBUG) header(NONE,"Debug");
   
   /*debugging*/	
   if (DEBUG) printf("main: Env-Variable lesen<br>\n");
   
	/*Die Benutzereingaben einlesen*/
   env=getenv("CONTENT_LENGTH");
   cl = env?atoi(env):0;
	if (cl == 0) error=NO_INPUT_STRING;
	
	/*debugging*/
	if (DEBUG) printf("main: cl: %d, Error: %d<br>",cl,error);
	
	/*Hier werden die Daten eingelesen und in einem Feld des typs "entry" abgelegt*/

	if (!error){
    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
		  if (DEBUG) printf("%d %s %s<br>",x,entries[x].name,entries[x].val);
    }
   }
   parameters=x;
   
   /*debugging*/
	if (DEBUG) printf("main: parameters: %d<br>",parameters);
	   
   /*Die Exporte einlesen*/
 	if (!error) error=read_printers(&printer_list);
 
	/*Die einzelnen Parameter herausnehmen und ueberpruefen*/
   /*action wird als erstes ausgewertet, da bei action=aktualisieren
     keine Defaultwerte ermittelt werden duerfen*/
   if (!error) {  
   	if (strcmp(entries[parameters-1].val,"")) strcpy(action,entries[parameters-1].val);
   	else error=NO_ACTION_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d Action: %s<br>",error,action);

	/*Der Name des Druckers*/
   if (!error) {
		strcpy(name,entries[0].val);
		backtoforward(name);
 		if (ContainsChr(name,':') || ContainsChr(name,'.') || ContainsChr(name,' ') || ContainsChr(name,'/')) error=INVALID_PRINTER_NAME;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d name: %s<br>",error,name);


	/*Der Typ des Druckers*/
   if (!error) {
		strcpy(type,entries[1].val);
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d type: %s<br>",error,type);


	/*Formfeed?*/
   if (!error) {
		if (!strcmp(entries[2].val,"ja")) formfeed=FALSE;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d formfeed: %d<br>",error,formfeed);

	/*Headerpage?*/
   if (!error) {
		if (!strcmp(entries[3].val,"ja")) headerpage=FALSE;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d headerpage: %d<br>",error,headerpage);

	/*Der Name des Druckservers*/
   if (!error) {
		strcpy(printserver,entries[4].val);
		backtoforward(printserver);
 		if (ContainsChr(printserver,':') || ContainsChr(printserver,' ') || ContainsChr(printserver,'/')) error=INVALID_PRINTSERVER_NAME;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d printserver: %s<br>",error,printserver);

	/*Der Name des Remote-Druckers*/
   if (!error) {
		strcpy(remotename,entries[5].val);
		backtoforward(remotename);
 		if (ContainsChr(remotename,':') || ContainsChr(remotename,'.') || ContainsChr(remotename,' ') || ContainsChr(remotename,'/')) error=INVALID_REMOTE_PRINTER_NAME;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d remotename: %s<br>",error,remotename);

	/*SMB-Passwort*/
   if (!error) {
   	if (!strcmp(entries[6].val,entries[7].val)) strcpy(rpasswd,entries[6].val);
   	else error=PASSWORDS_DIFFERENT;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d Password: %s<br>",error,rpasswd);

	/*Limit*/
   if (!error) {
		if (isNumber(entries[8].val)) max=atoi(entries[6].val);
		else	if (!strlen(entries[8].val)) {
					if (!strcmp(action,"aktualisieren")) max=-1;
					else max=0;
				} else error=MAX_NOT_NUMBER;	
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d max: %d<br>",error,max);

	/*debugging*/
	if (DEBUG) printf("Vor der Aktion: Error: %d<br>",error);
	
   		
   /*Hier wird ausgewaehlt, was der User will*/
	
   if (!error){
		if (!strcmp(action,"eintragen")) { 
			if (DEBUG) printf("main: add_printer aufrufen<br>");
			error=add_printer(printer_list, 
			                  name, type, printserver, remotename, formfeed, headerpage, max, rpasswd);
		}
		else { 
			if (!strcmp(action,"entfernen") && !error) {
				if (DEBUG) printf("main: delete_printer aufrufen<br>");
					error=delete_printer(printer_list, 
			                  name, type, printserver, remotename, formfeed, headerpage, max, rpasswd);
			}
			else { 
				if (DEBUG) printf("main: change_printer aufrufen<br>");
				if (!strcmp(action,"aktualisieren") && !error) 	
					error=change_printer(printer_list, 
			                  name, type, printserver, remotename, formfeed, headerpage, max, rpasswd);

			}
		}				
   }
   

	/*debugging*/
	printf("Nach der Aktion: Error: %d<br>",error);
	if (DEBUG) tail();

	if (!DEBUG) {
		if (error == NO_ERROR) {
			header(NONE,"Erfolgreich!");
			printf("Die Aktion wurde erfolgreich abgeschlossen\n");
			tail();
		}
		else
			{
				switch (error) {
					case QUERY_STRING_NOT_SET : print_error("Es wurden keine Parameter empfangen ($QUERY_STRING nicht gesetzt).");
					     break;
					case USERID_NOT_NUMBER : print_error("Die Benutzernummer ist keine Zahl");
					     break;
					case GROUP_NOT_NUMBER : print_error("Die Gruppennummer ist keine Zahl");
					     break;
					case USER_EXISTS : print_error("Der Benutzer existiert bereits");
				   	  break;
					case USER_UNKNOWN : print_error("Der Benutzer ist unbekannt");
					     break;
					case USERID_EXISTS : print_error("Die Benutzernummer ist bereits vergeben");
					     break;
					case USERID_UNKNOWN : print_error("Die Benutzernummer ist unbekannt");
				   	  break;
					case GROUP_EXISTS : print_error("Die Gruppe existiert bereits");
					     break;
					case GROUP_UNKNOWN : print_error("Die Gruppe ist unbekannt");
					     break;
					case PASSWD_FAILURE : print_error("Die Datei /etc/passwd konnte nicht gelesen/geschrieben werden");
				   	  break;
					case GROUP_FAILURE : print_error("Die Datei /etc/group konnte nicht gelesen/geschrieben werden");
					     break;
					case NO_GROUPNAME_GIVEN : print_error("Sie m&uuml;ssen einen Gruppennamen angeben!");
					     break;
					case NO_ACTION_GIVEN : print_error("Sie m&uuml;ssen hinzuf&uuml;gen, l&ouml;schen oder &auml;ndern ausw&auml;hlen!");
				   	  break;
					case NO_INPUT_STRING : print_error("Es wurden keine Eingaben aus dem Browser empfangen!");
					     break;
					case PASSWORDS_DIFFERENT : print_error("Die eingegebenen Passw&ouml;rter stimmen nicht &uuml;berein.");
					     break;
					case GROUPID_EXISTS : print_error("Diese Gruppennummer existiert bereits.");
					     break;
					case GROUPID_UNKNOWN : print_error("Diese Gruppennummer ist unbekannt.");
					     break;
					case GROUP_IS_HOME : print_error("Die Gruppe ist die Heimatgruppe eines Benutzers, bitte &auml;ndern sie dessen Heimatgruppe bevor sie die Gruppe l&ouml;schen.");
					     break;
					case FILESYS_FAILURE : print_error("Die Filesystemtabelle konnte nicht geladen werden");
						  break;
					case IMPORT_FAILURE : print_error("Die Import-Tabelle konnte nicht geladen werden");
						  break;
					case NO_DRIVE_GIVEN : print_error("Sie haben vergessen einen Partitionsbuchstaben anzugeben");
						  break;
					case NO_VALID_MOUNTPOINT : print_error("Sie haben ein fehlerhaftes Ziel angegeben (ev. ist das Ziel schon vergeben)");
						  break;
					case DRIVE_ALREADY_MOUNTED : print_error("Dieses Laufwerk ist schon angemeldet");
						  break;
					case DRIVE_UNKNOWN : print_error("Dieses Laufwerk ist im System unbekannt");
						  break;
					case IMPORT_ALREADY_MOUNTED : print_error("Dieser Importname ist bereits vergeben und die betreffende Verbindung ist aktiv");
						  break;
					case IMPORT_UNKNOWN : print_error("Dieser Importname ist im System unbekannt");
						  break;
					case MOUNT_FAILED : print_error("Die Verbindung konnte nicht erstellt werden, &uuml;berpr&uuml;fen sie bitte die Parameter und &auml;ndern sie"
					                                " diese gegebenenfalls.");
						  break;
					case NO_SLASHES_ALLOWED : print_error("Der Name des Importes darf keine \" / \" oder \" \\ \" enthalten");
						  break;
					case PATH_UNKNOWN : print_error("Der Exportpfad konnte nicht gefunden werden.");
						  break;
					case SMB_NAME_EXISTS : print_error("Dieser Windows-Name ist bereits vergeben.");
						  break;
					case EXPORT_EXISTS : print_error("Dieser Pfad ist bereits exportiert.");
						  break;
					case EXPORT_UNKNOWN : print_error("Dieser Pfad ist nicht exportiert.");
						  break;
					case SMB_NAME_NOT_SET : print_error("Wenn ein Verzeichnis Windows-kompatibel exportiert werden soll, muss ein Windows-Name angegeben werden.");
						  break;
					case  INVALID_PRINTER_NAME: print_error("Der Druckername enth&auml;lt ung&uuml;ltige Zeichen.");
						  break;
					case  INVALID_PRINTSERVER_NAME: print_error("Der Name des Druckservers enth&auml;lt ung&uuml;ltige Zeichen.");
						  break;
					case  INVALID_REMOTE_PRINTER_NAME: print_error("Der Name des Netzdruckers enth&auml;lt ung&uuml;ltige Zeichen.");
						  break;
					case  PRINTER_ALREADY_EXISTS: print_error("Der Druckername ist bereits vergeben.");
						  break;
					case  SMBCONF_FAILURE: print_error("Die Konfigurationsdatei f&uuml;r den SMB-Drucker konnte nicht erstellt werden.");
						  break;
					case  PRINTER_UNKNOWN: print_error("Dieser Drucker ist unbekannt.");
						  break;
					case  PRINTSERVER_MISSING: print_error("Der Name des Rechners, an den der Drucker angeschlossen ist fehlt.");
						  break;
					case  REMOTENAME_MISSING: print_error("Der Name des Druckers auf dem Druckserver fehlt.");
						  break;
					case  MAX_NOT_NUMBER: print_error("Das Druckerlimit muss eine Zahl sein. Lassen sie das Feld leer oder geben sie eine Null ein, wenn kein Limit eingerichtet werden soll.");
						  break;
				}
			}
	}
	return 0;
}
