/******************************************************************************
* Aendern von /etc/fstab																	   *
* 																										*
******************************************************************************/

/* Die Includes werden gemacht */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <support.h>
#include <string.h>
#include <ctype.h>
#include "util.h"



/*Lokale Defines */
/*---------------------------------------------------------------------------*/

/*in print-error wird eine Fehlerseite generiert und ausgegeben*/
/*---------------------------------------------------------------------------*/
void print_error(char* error){

	header(NONE,"Fehler");
	printf("%s\n",error);
	tail();
}


/*in add_filesys wird ein Dateisystem in fstab eingetragen*/
/*---------------------------------------------------------------------------*/
int add_filesys(struct structUser *user_base, 
             struct structGroup *group_base,
             struct structFilesys *filesys_base,
             struct structImport *import_base,
             char drive[1],
             char	mountpoint[MAXLINE],
             char type[10],
             char owner[MAXLINE],
             char	group[MAXLINE],
             int	owner_perm,
             int	group_perm,
             int	world_perm) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE					*passwd_file;			/*File-Handle um in /etc/passwd zu
	                                         schreiben*/
	FILE					*group_file;			/*File-Handle um in /etc/group zu
														  schreiben*/
	struct structFilesys *akt;
	char					*buffer;
	char					permissions[4];

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (DEBUG) printf("add_filesys: Drive: %s<br>",drive);
	if (FilesysExistsByPart(filesys_base,drive)) 
		return DRIVE_ALREADY_MOUNTED;

	akt = (struct structFilesys *) malloc(sizeof(struct structFilesys));
	strcpy(akt->drive,drive);
	strcpy(akt->mountpoint,mountpoint);
	strcpy(akt->type,type);
	strcpy(akt->owner,owner);
	strcpy(akt->group,group);
	if (owner_perm+group_perm+world_perm > 0) 
		sprintf(permissions, "%d%d%d", owner_perm, group_perm, world_perm);
	else 
		strcpy(permissions,"-");
	
	strcpy(akt->permissions,permissions);
	addFilesys(&filesys_base,akt);
	FlushFstab(filesys_base,import_base);
}


/*in delete_filesys wird ein Filesys aus der fstab geloescht*/
/*---------------------------------------------------------------------------*/
int delete_filesys(struct structUser *user_base, 
             struct structGroup *group_base,
             struct structFilesys *filesys_base,
             struct structImport *import_base,
             char drive[1],
             char	mountpoint[MAXLINE],
             char type[10],
             char owner[MAXLINE],
             char	group[MAXLINE],
             int	owner_perm,
             int	group_perm,
             int	world_perm) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE					*passwd_file;			/*File-Handle um in /etc/passwd zu
	                                         schreiben*/
	FILE					*group_file;			/*File-Handle um in /etc/group zu
														  schreiben*/
	struct structFilesys *akt;
	char					*buffer;
	char					permissions[4];

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (!FilesysExistsByPart(filesys_base,drive)) return DRIVE_UNKNOWN;
	if (DEBUG) printf("delete_filesys: Aufruf von delFilesys()<br>");
	delFilesys(&filesys_base,drive);
	if (DEBUG) printf("delete_filesys: Aufruf von FlushFstab()<br>");
	FlushFstab(filesys_base,import_base);
}


/*mit change_filesys koennen die mount-Parameter veraendert werden*/
/*---------------------------------------------------------------------------*/
int change_filesys(struct structUser *user_base, 
             struct structGroup *group_base,
             struct structFilesys *filesys_base,
             struct structImport *import_base,
             char drive[1],
             char	mountpoint[MAXLINE],
             char type[10],
             char owner[MAXLINE],
             char	group[MAXLINE],
             int	owner_perm,
             int	group_perm,
             int	world_perm) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE					*passwd_file;			/*File-Handle um in /etc/passwd zu
	                                         schreiben*/
	FILE					*group_file;			/*File-Handle um in /etc/group zu
														  schreiben*/
	struct structFilesys *akt;
	char					*buffer;
	char					permissions[4];

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	if (!FilesysExistsByPart(filesys_base,drive)) return DRIVE_UNKNOWN;
	akt = GetFilesys(filesys_base,drive);
	if (DEBUG) printf("change_filesys: mountpoint: %s, type: %s,owner: %s,group: %s<br>",
							mountpoint,type,owner,group);
	if (DEBUG) printf("change_filesys: akt->mountpoint: %s, akt->type: %s, akt->owner: %s, akt->group: %s<br>",
							akt->mountpoint,akt->type,akt->owner,akt->group);
	if (strcmp(mountpoint,"")) strcpy(akt->mountpoint,mountpoint);
	if (strcmp(type,"")) strcpy(akt->type,type);	
	if (strcmp(owner,"")) strcpy(akt->owner,owner);
	if (strcmp(group,"")) strcpy(akt->group,group);
	if (owner_perm+group_perm+world_perm > 0) {
		sprintf(permissions, "%d%d%d", owner_perm, group_perm, world_perm);
		if (DEBUG) printf("change_filesys: owner: %d, group: %d, world: %d, permissions: %s<br>",owner_perm,group_perm,world_perm,permissions);	
		strcpy(akt->permissions,permissions);
	}
	FlushFstab(filesys_base,import_base);
}



/* In main wird die HTML-Ausgabe initialisiert und alles gesteuert */
/*---------------------------------------------------------------------------*/

void main()
{
	struct structUser 	*user_list=NULL;		/*Anker der Userliste */
	struct structGroup 	*group_list=NULL;		/*Anker der Gruppenliste */
	struct structFilesys	*filesys_list=NULL;	/*Anker der Filesystemliste */
	struct structImport	*import_list=NULL;	/*Anker der Importliste */
	char						*env;						/*Parameterstring der vom Web-Server
	                                       	  kommt*/
	int						error=0;					/*Flag das eine Fehlernummer 
															  enthaelt*/
	char						dummy_str[MAXLINE]; 	/*Fuer String-Ergebnisse, die 
															  uninteressant sind*/
	unsigned char			drive[2];
	char						mountpoint[MAXLINE];
	char						owner[MAXLINE];
	char						group[MAXLINE];
	int						owner_perm=0,owner_r=0,owner_w=0,owner_x=0;						
	int						group_perm=0,group_r=0,group_w=0,group_x=0;						
	int						world_perm=0,world_r=0,world_w=0,world_x=0;						
	char						type[10];
	char						action[32];
	entry						entries[20];
	int						cl;
	int						x,m;
	int						parameters,test;
	char						dummy_chr;



	/*debugging*/
	if (DEBUG) header(NONE,"Debug");
   
   /*debugging*/	
   if (DEBUG) printf("main: Env-Variable lesen<br>\n");
   
	/*Die Benutzereingaben einlesen*/
   env=getenv("CONTENT_LENGTH");
   cl = env?atoi(env):0;
	if (cl == 0) error=NO_INPUT_STRING;
	
	/*debugging*/
	if (DEBUG) printf("main: cl: %d, Error: %d<br>",cl,error);
	
	/*Hier werden die Daten eingelesen und in einem Feld des typs "entry" abgelegt*/

	if (!error){
    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
		  if (DEBUG) printf("%d %s %s<br>",x,entries[x].name,entries[x].val);
    }
   }
   parameters=x;
   
   /*debugging*/
	if (DEBUG) printf("main: parameters: %d<br>",parameters);
	   
	/*Die passwd-Datei einlesen*/
	if (!error) error=read_users(&user_list);
   
      
   /*Die group-Datei einlesen*/
 	if (!error) error=read_groups(&group_list);

   /*Die filesysteme einlesen*/
 	if (!error) error=read_filesys(&filesys_list);

   /*Die imports einlesen*/
 	if (!error) error=read_imports(&import_list);

	/*Die einzelnen Parameter herausnehmen und ueberpruefen*/
   /*action wird als erstes ausgewertet, da bei action=aktualisieren
     keine Defaultwerte ermittelt werden duerfen*/
   if (!error) {  
   	if (strcmp(entries[parameters-1].val,"")) strcpy(action,entries[parameters-1].val);
   	else error=NO_ACTION_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d Action: %s<br>",error,action);


	/*Der Partitionsbuchstabe wird gelesen*/
   if (!error) {
		strcpy(drive,entries[0].val);
		if (!strcmp(drive,"")) error=NO_DRIVE_GIVEN;
		else 	if (drive[0] < 'a') drive[0] += 32;

   }
	/*debugging*/
	if (DEBUG) printf("main: Error: %d Partitionsbuchstabe: %s<br>",error,drive);

	if (!error) {
	   
	   /*Der Mountpoint wird ausgewertet, wenn keiner angegeben ist, und das Filesystem
	     neu eingetragen werden soll, dann wird ein default-mntpt erzeugt.*/
		strcpy(mountpoint,entries[1].val);
		if ((!strcmp(mountpoint,"")) && (strcmp(action,"aktualisieren"))) strcpy(mountpoint,"/");
		if (!IsValidMountpoint(filesys_list,import_list,mountpoint) && (strcmp(action,"aktualisieren"))) 
			error=NO_VALID_MOUNTPOINT;
      /*debugging*/
		if (DEBUG) printf("main: Error: %d IsValidMountpoint: %d mountpoint: %s<br>",
								error,IsValidMountpoint(filesys_list,import_list,mountpoint),mountpoint);


	   /*Der Typ wird ausgewertet*/
		strcpy(type,entries[parameters-2].val);
      /*debugging*/
		if (DEBUG) printf("main: Error: %d type: %s<br>",error,type);
	

	   /*Der Owner wird ausgewertet*/
		strcpy(owner,entries[2].val);
		if ((!strcmp(owner,"")) && (strcmp(action,"aktualisieren"))) {
			if (!strcmp(type,"minix")) strcpy(owner,"-");
			else strcpy(owner,"root");
		}
		if (!UserExists(user_list,owner) 
			&& (strcmp(action,"aktualisieren"))
			&& (strcmp(owner,"-"))) error=USER_UNKNOWN;
      /*debugging*/
		if (DEBUG) printf("main: Error: %d owner: %s<br>",error,owner);


	   /*Die Group wird ausgewertet*/
		strcpy(group,entries[3].val);
		if ((!strcmp(group,"")) && (strcmp(action,"aktualisieren"))) {
			if (!strcmp(type,"minix")) strcpy(group,"-");
			else strcpy(group,"wheel");
		}

		if (!strcmp(action,"aktualisieren")) {
			if (DEBUG) printf("main: Gruppe auswerten, action ist aktualisieren<br>\n");			
			if (!GnameExists(group_list,group) && strcmp(group,"")) 
				error=GROUP_UNKNOWN;
		}
		else {
			if (DEBUG) printf("main: Gruppe auswerten, action ist nicht aktualisieren<br>\n");			
			if (!GnameExists(group_list,group) && strcmp(group,"-")) 
				error=GROUP_UNKNOWN;
		}


      /*debugging*/
		if (DEBUG) printf("main: Error: %d group: %s<br>",error,group);

	   /*Die Berechtigungen werden ausgewertet*/

		x=0;
		while (strcmp(entries[x+4].name,"filesys")) {
			if (!strcmp(entries[x+4].val,"owner_r")) owner_r=4;
			if (!strcmp(entries[x+4].val,"owner_w")) owner_w=2;
			if (!strcmp(entries[x+4].val,"owner_x")) owner_x=1;
			if (!strcmp(entries[x+4].val,"group_r")) group_r=4;
			if (!strcmp(entries[x+4].val,"group_w")) group_w=2;
			if (!strcmp(entries[x+4].val,"group_x")) group_x=1;
			if (!strcmp(entries[x+4].val,"world_r")) world_r=4;
			if (!strcmp(entries[x+4].val,"world_w")) world_w=2;
			if (!strcmp(entries[x+4].val,"world_x")) world_x=1;
			x++;
		}
	
		owner_perm=owner_r+owner_w+owner_x;
		group_perm=group_r+group_w+group_x;
		world_perm=world_r+world_w+world_x;
      if ((world_perm+group_perm+owner_perm == 0) && (strcmp(action,"aktualisieren"))) {
			if (!strcmp(type,"dos")) {
				owner_perm=7;
				group_perm=7;
				world_perm=7;
			}
      }
      /*debugging*/
		if (DEBUG) printf("main: Error: %d Perms: %d%d%d<br>",error,owner_perm,group_perm,world_perm);

	}
	/*debugging*/
	if (DEBUG) printf("Vor der Aktion: Error: %d<br>",error);
	
		
   /*Hier wird ausgewaehlt, was der User will*/
   if (!error){
		if (!strcmp(action,"eintragen")) { 
			if (DEBUG) printf("main: add_filesys aufrufen<br>");
			error=add_filesys(user_list, group_list, filesys_list, import_list, 
			                  drive, mountpoint, type, owner, group, owner_perm, group_perm, world_perm);
		}
		else { 
			if (!strcmp(action,"entfernen") && !error) {
				if (DEBUG) printf("main: delete_filesys aufrufen<br>");
				error=delete_filesys(user_list, group_list, filesys_list, import_list, 
			   	               drive, mountpoint, type, owner, group, owner_perm, group_perm, world_perm);
			}
			else { 
				if (DEBUG) printf("main: change_filesys aufrufen<br>");
				if (!strcmp(action,"aktualisieren") && !error) 	error=change_filesys(user_list, group_list, filesys_list, import_list, 
			                  																		drive, mountpoint, type, owner, group, owner_perm, group_perm, world_perm);

			}
		}				
   }
   

	/*debugging*/
	printf("Nach der Aktion: Error: %d<br>",error);
	if (DEBUG) tail();

	if (!DEBUG) {
		if (error == NO_ERROR) {
			header(NONE,"Erfolgreich!");
			printf("Die Aktion wurde erfolgreich abgeschlossen\n");
			tail();
		}
		else
			{
				switch (error) {
					case QUERY_STRING_NOT_SET : print_error("Es wurden keine Parameter empfangen ($QUERY_STRING nicht gesetzt).");
					     break;
					case USERID_NOT_NUMBER : print_error("Die Benutzernummer ist keine Zahl");
					     break;
					case GROUP_NOT_NUMBER : print_error("Die Gruppennummer ist keine Zahl");
					     break;
					case USER_EXISTS : print_error("Der Benutzer existiert bereits");
				   	  break;
					case USER_UNKNOWN : print_error("Der Benutzer ist unbekannt");
					     break;
					case USERID_EXISTS : print_error("Die Benutzernummer ist bereits vergeben");
					     break;
					case USERID_UNKNOWN : print_error("Die Benutzernummer ist unbekannt");
				   	  break;
					case GROUP_EXISTS : print_error("Die Gruppe existiert bereits");
					     break;
					case GROUP_UNKNOWN : print_error("Die Gruppe ist unbekannt");
					     break;
					case PASSWD_FAILURE : print_error("Die Datei /etc/passwd konnte nicht gelesen/geschrieben werden");
				   	  break;
					case GROUP_FAILURE : print_error("Die Datei /etc/group konnte nicht gelesen/geschrieben werden");
					     break;
					case NO_GROUPNAME_GIVEN : print_error("Sie m&uuml;ssen einen Gruppennamen angeben!");
					     break;
					case NO_ACTION_GIVEN : print_error("Sie m&uuml;ssen hinzuf&uuml;gen, l&ouml;schen oder &auml;ndern ausw&auml;hlen!");
				   	  break;
					case NO_INPUT_STRING : print_error("Es wurden keine Eingaben aus dem Browser empfangen!");
					     break;
					case PASSWORDS_DIFFERENT : print_error("Die eingegebenen Passw&ouml;rter stimmen nicht &uuml;berein.");
					     break;
					case GROUPID_EXISTS : print_error("Diese Gruppennummer existiert bereits.");
					     break;
					case GROUPID_UNKNOWN : print_error("Diese Gruppennummer ist unbekannt.");
					     break;
					case GROUP_IS_HOME : print_error("Die Gruppe ist die Heimatgruppe eines Benutzers, bitte &auml;ndern sie dessen Heimatgruppe bevor sie die Gruppe l&ouml;schen.");
					     break;
					case FILESYS_FAILURE : print_error("Die Filesystemtabelle konnte nicht geladen werden");
						  break;
					case IMPORT_FAILURE : print_error("Die Import-Tabelle konnte nicht geladen werden");
						  break;
					case NO_DRIVE_GIVEN : print_error("Sie haben vergessen einen Partitionsbuchstaben anzugeben");
						  break;
					case NO_VALID_MOUNTPOINT : print_error("Sie haben ein fehlerhaftes Ziel angegeben (ev. ist das Ziel schon vergeben)");
						  break;
					case DRIVE_ALREADY_MOUNTED : print_error("Dieses Laufwerk ist schon angemeldet");
						  break;
					case DRIVE_UNKNOWN : print_error("Dieses Laufwerk ist im System unbekannt");
						  break;
				}
			}
	}
}
