/******************************************************************************
* Dieses Programm dient zum aendern der Benutzerdaten (/etc/passwd) auf einem *
* Unix-Rechner																						*
******************************************************************************/

/* Die Includes werden gemacht */
/*---------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <support.h>
#include <string.h>
#include "util.h"



/*Lokale Defines */
/*---------------------------------------------------------------------------*/

/*in print-error wird eine Fehlerseite generiert und ausgegeben*/
/*---------------------------------------------------------------------------*/
void print_error(char* error){

	header(NONE,"Fehler");
	printf("%s<br>",error);
	tail();
}


/*in add_user wird ein Benutzer eingefuegt*/
/*---------------------------------------------------------------------------*/
int add_user(struct structUser *user_base, 
             struct structGroup *group_base,
             char name[UNAME_LEN],
             char password[PASSWORD_LEN],
	          long int userid,
	          long int group,
	          char fullname[MAXLINE],
	          char shell[MAXLINE]) {


	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE					*passwd_file;			/*File-Handle um in /etc/passwd zu
	                                         schreiben*/
	FILE					*group_file;			/*File-Handle um in /etc/group zu
														  schreiben*/
	struct structUser *new_user;				/*Die Daten des neuen Users*/
	struct structUser *prev;
	struct structUser *next;
	struct structUser *akt;
	struct structGroup *akt_group;
	char					homedir[MAXLINE];		/*Das Homedir wird automatisch
	                                         erzeugt*/
	char					shell_path[MAXLINE]; /*Der Pfad zur Loginshell wird
	                                         automatisch erzeugt*/
	char					*dummy_str;
	char					*createhome;			/*Befehlsstring um das Home zu 
	                                         erzeugen*/
	char					groupname[MAXLINE];	/*Der Name der Gruppe*/
	long int					akt_uid;


   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   
   /*Die Parameter auf Korrektheit pruefen*/
	if (UserExists(user_base, name)) return USER_EXISTS;
	if (UidExists(user_base, userid)) return USERID_EXISTS;
	if (!GidExists(group_base, group)) return GROUP_UNKNOWN; 
		
	/*Der neue User wird in die Userliste eingefuegt*/
	new_user = (struct structUser *) malloc(sizeof(struct structUser));
	strcpy(new_user->username,name);
	strcpy(new_user->password,password);
	new_user->uid=userid;
	new_user->gid=group;
	strcpy(new_user->gecos,fullname);
	/*Erstellen des Homedir-Pfades*/
	strcpy(homedir,"/home/");
	dummy_str=strcat(homedir,name);
	strcpy(new_user->homedir,homedir);
	/*Erstellen des shell_paths*/
	strcpy(shell_path,"/bin/");
	dummy_str=strcat(shell_path,shell);
	strcpy(new_user->shell,shell_path);

	akt=user_base;
	while ((akt->uid <= new_user->uid) && (akt->next != user_base)) {
		akt=akt->next;
	}
	next=akt;
   prev=akt->prev;
	prev->next=new_user;
	next->prev=new_user;
	new_user->next=next;
	new_user->prev=prev;	

	/*Der neue User wird in die Gruppenliste eingefuegt*/
	dummy_str=(char*) malloc(sizeof(char)*MAXLINE);
	akt_group=group_base;
	while ((akt_group->gid != new_user->gid) && (akt_group->next != group_base)) 
		akt_group=akt_group->next;
	if (akt_group->gid == new_user->gid) {
		addUserToGroup(akt_group,new_user->username);		
		strcpy(groupname,akt_group->groupname);
	}
	else return GROUP_UNKNOWN;

	/*/etc/group wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/group /etc/group.old");

	if (!(group_file=fopen(GROUP_NAME,"wb"))) {
		return GROUP_FAILURE;
	}

	/*Die Gruppenliste wird geschrieben*/
	akt_group=group_base;
	do{
		fprintf(group_file,"%s:*:%d:%s\n",akt_group->groupname,akt_group->gid,akt_group->users);
		akt_group=akt_group->next;
	}while (akt_group != group_base);
	fclose(group_file);

	
	/*/etc/passwd wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/passwd /etc/passwd.old");

	if (!(passwd_file=fopen(PASSWD_NAME,"wb"))) {
		return PASSWD_FAILURE;
	}

	/*Die Userliste wird geschrieben*/
	akt=user_base;
	do{
		fprintf(passwd_file,"%s:%s:%d:%d:%s:%s:%s\n",akt->username,akt->password,akt->uid,akt->gid,akt->gecos,akt->homedir,akt->shell);
		akt=akt->next;
	}while (akt != user_base);
	fclose(passwd_file);
	
	/*Das Home wird erzeugt*/
	createhome = (char*) malloc(sizeof(char)*MAXLINE);
	sprintf(createhome,"cp -r /home/.default %s\n",new_user->homedir);
	system(createhome);
	if (DEBUG) printf("%s<br>",createhome);
	sprintf(createhome,"chown -R %s %s\n",new_user->username,new_user->homedir);
	system(createhome);
	if (DEBUG) printf("%s<br>",createhome);
	sprintf(createhome,"chgrp -R %d %s\n",new_user->gid,new_user->homedir);
	system(createhome);
	sprintf(createhome,"chmod 755 %s\n",new_user->homedir);
	system(createhome);
	if (DEBUG) printf("%s<br>",createhome);
	return 0;
	free(dummy_str);
}


/*in delete_user wird ein Benutzer geloescht*/
/*---------------------------------------------------------------------------*/
int delete_user(struct structUser *user_base, 
             struct structGroup *group_base,
             char name[UNAME_LEN],
             char password[PASSWORD_LEN],
	          long int userid,
	          long int group,
	          char fullname[MAXLINE],
	          char shell[MAXLINE]) {
	          
	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE						*passwd_file;	/*Filehandle fuer /etc/passwd*/
	FILE						*group_file;	/*Filehandle fuer /etc/group*/
	struct structUser 	*prev;			/*Pointer zum Manipulieren*/
	struct structUser 	*next;         /*der Benutzerliste*/
	struct structUser 	*akt;
	struct structGroup	*akt_group;    /*Pointer zum Manipulieren der*/
	                                    /*Gruppenliste*/
   char						*delhome;		/*Hier wird der Befehlsstring zum*/
   												/*loeschen des Homes generiert*/	
   
   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/

	/*Wenn der User nicht existiert kann er nicht geloescht werden*/
	if (!UserExists(user_base, name)) return USER_UNKNOWN;
	
	/*Es wird nach dem Benutzer gesucht*/
	akt=user_base;
	while(strcmp(akt->username,name)) akt=akt->next;
	
	/*Der Benutzer wird aus der Userliste entfernt*/
	prev=akt->prev;
	next=akt->next;
	prev->next=next;
	next->prev=prev;

	/*Das Homedirectory des Benutzers wird geloescht*/
	delhome=(char*) malloc(sizeof(char)*MAXLINE);
	sprintf(delhome,"rm -r %s",akt->homedir);
	system(delhome);
	free(delhome);
	/*Freigeben des Speichers des geloeschten Benutzers*/
	free(akt);

	/*Der User wird aus der Gruppenliste entfernt*/
	akt_group=group_base;
	do{
		if (DEBUG) printf("delete_user: Gruppe: %s, Liste: %s<br>",akt_group->groupname,akt_group->users);
		if (UserInGroup(akt_group,name)) delUserFromGroup(akt_group,name);
		if (DEBUG) printf("delete_user: Gruppe: %s, Liste: %s<br>",akt_group->groupname,akt_group->users);
		akt_group=akt_group->next;
	}while(akt_group->next != group_base);

	/*/etc/group wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/group /etc/group.old");

	if (!(group_file=fopen(GROUP_NAME,"wb"))) {
		return GROUP_FAILURE;
	}

	/*Die Gruppenliste wird geschrieben*/
	akt_group=group_base;
	do{
		fprintf(group_file,"%s:*:%d:%s\n",akt_group->groupname,akt_group->gid,akt_group->users);
		akt_group=akt_group->next;
	}while (akt_group != group_base);
	fclose(group_file);

	
	/*/etc/passwd wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/passwd /etc/passwd.old");

	if (!(passwd_file=fopen(PASSWD_NAME,"wb"))) {
		return PASSWD_FAILURE;
	}

	/*Die Userliste wird geschrieben*/
	akt=user_base;
	do{
		fprintf(passwd_file,"%s:%s:%d:%d:%s:%s:%s\n",akt->username,akt->password,akt->uid,akt->gid,akt->gecos,akt->homedir,akt->shell);
		akt=akt->next;
	}while (akt != user_base);
	fclose(passwd_file);
	
	/*debugging*/
	if (DEBUG) printf("Ende von delete_user() erreicht<br>");
	return 0;
}


/*mit change_user koennen Benutzerdaten veraendert werden*/
/*---------------------------------------------------------------------------*/
int change_user(struct structUser *user_base, 
             struct structGroup *group_base,
             char name[UNAME_LEN],
             char password[PASSWORD_LEN],
	          long int userid,
	          long int group,
	          char fullname[MAXLINE],
	          char shell[MAXLINE]) {

	/*Variablen*/
   /*-----------------------------------------------------------------------*/
	FILE					*passwd_file;			/*File-Handle um in /etc/passwd zu
	                                         schreiben*/
	FILE					*group_file;			/*File-Handle um in /etc/group zu
														  schreiben*/
	struct structUser *prev;
	struct structUser *next;
	struct structUser *akt;
	struct structGroup *akt_group;
	char					homedir[MAXLINE];		/*Das Homedir wird automatisch
	                                         erzeugt*/
	char					shell_path[MAXLINE]; /*Der Pfad zur Loginshell wird
	                                         automatisch erzeugt*/
	char					*dummy_str;
	char					*buffer;
	char					*createhome;			/*Befehlsstring um das Home zu 
	                                         erzeugen*/
	char					groupname[MAXLINE];	/*Der Name der Gruppe*/
	char					*name_old_group;
	long int				akt_uid;
	long int				old_group;

   /*Prozedurkoerper*/
   /*-----------------------------------------------------------------------*/
   
	/*debugging*/
	if (DEBUG) printf("Vor der Parameterpruefung<br>");
   
   /*Die Parameter auf Korrektheit pruefen*/
	if (!UserExists(user_base, name)) return USER_UNKNOWN;
	if (UidExists(user_base, userid)) return USERID_EXISTS;
	if ((group != -1) && (!GidExists(group_base, group))) return GROUP_UNKNOWN; 
	
	/*debugging*/
	if (DEBUG) printf("Nach der Parameterpruefung<br>");
		
	/*Der User wird gesucht*/
	akt=user_base;
	while (strcmp(akt->username,name)) akt=akt->next;
	
	/*debugging*/
	if (DEBUG) printf("Benutzer gefunden<br>");

	/*Die alte Gruppe wird gespeichert*/
	old_group=akt->gid;
	
	/*Die gewuenschten Aenderungen werden geschrieben*/
	if (strcmp(password,"")) strcpy(akt->password,password);
	if (userid != -1) akt->uid=userid;
	if (group != -1) akt->gid=group;
	if (strcmp(fullname,"")) strcpy(akt->gecos,fullname);
	/*Erstellen des shell_paths*/
	strcpy(shell_path,"/bin/");
	dummy_str=strcat(shell_path,shell);
	if (strcmp(shell,"default")) strcpy(akt->shell,shell_path);

	if (DEBUG) {
		akt_group=group_base;
		do {
			if (DEBUG) printf("change_user: akt_group->groupname: %s, akt_group->gid: %d, akt_group->users: %s<br>",akt_group->groupname,akt_group->gid,akt_group->users);
			akt_group=akt_group->next;
		} while(akt_group != group_base);
	}

	/*Wenn der User zu einer neuen Gruppe gehoeren soll wird dies erledigt*/
	if (group != -1) {
		/*debugging*/
		if (DEBUG) printf("change_user: Gruppen&auml;nderung auf %d<br>",akt->gid);
		/*Der User wird aus seiner alten Gruppe geloescht*/
		if (DEBUG) printf("change_user: Old GroupId: %d<br>",old_group);
		
		akt_group=group_base;
		do{
			if (akt_group->gid != old_group) akt_group=akt_group->next;
		}while ((akt_group != group_base) && (akt_group->gid != old_group));
		delUserFromGroup(akt_group,akt->username);

		if (DEBUG) {
			akt_group=group_base;
			do {
				if (DEBUG) printf("change_user: akt_group->groupname: %s, akt_group->gid: %d, akt_group->users: %s<br>",akt_group->groupname,akt_group->gid,akt_group->users);
				akt_group=akt_group->next;
			} while(akt_group != group_base);
		}

		
		/*...und in seine neue Gruppe eingetragen*/
		akt_group=group_base;
		do {
			if (akt_group->gid != akt->gid) akt_group=akt_group->next;
			if (DEBUG) printf("change_user: akt->gid: %d, akt_group->gid: %d<br>",akt->gid,akt_group->gid);
		} while ((akt_group->gid != akt->gid) && (akt_group != group_base)); 
		if (DEBUG) printf("change_user: akt->gid: %d, akt_group->gid: %d<br>",akt->gid,akt_group->gid);
		if (akt_group->gid == akt->gid) {
			addUserToGroup(akt_group,akt->username);
			strcpy(groupname,akt_group->groupname);
		}
		else return GROUP_UNKNOWN;
   }
	
	
	/*/etc/group wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/group /etc/group.old");

	if (!(group_file=fopen(GROUP_NAME,"wb"))) {
		return GROUP_FAILURE;
	}

	/*Die Gruppenliste wird geschrieben*/
	akt_group=group_base;
	do{
		fprintf(group_file,"%s:*:%d:%s\n",akt_group->groupname,akt_group->gid,akt_group->users);
		akt_group=akt_group->next;
	}while (akt_group != group_base);
	fclose(group_file);

	
	/*/etc/passwd wird gesichert und zum schreiben geoeffnet*/
	system("mv /etc/passwd /etc/passwd.old");

	if (!(passwd_file=fopen(PASSWD_NAME,"wb"))) {
		return PASSWD_FAILURE;
	}

	/*Die Userliste wird geschrieben*/
	akt=user_base;
	do{
		fprintf(passwd_file,"%s:%s:%d:%d:%s:%s:%s\n",akt->username,akt->password,akt->uid,akt->gid,akt->gecos,akt->homedir,akt->shell);
		akt=akt->next;
	}while (akt != user_base);
	fclose(passwd_file);
	return 0;
}



/* In main wird die HTML-Ausgabe initialisiert und alles gesteuert */
/*---------------------------------------------------------------------------*/

void main()
{
	struct structUser 	*user_list=NULL;		/*Anker der Userliste */
	struct structGroup 	*group_list=NULL;		/*Anker der Gruppenliste */
	char						*env;						/*Parameterstring der vom Web-Server
	                                       	  kommt*/
	int						error=0;					/*Flag das eine Fehlernummer 
															  enthaelt*/
	char						dummy_str[MAXLINE]; 	/*Fuer String-Ergebnisse, die 
															  uninteressant sind*/
	char						username[UNAME_LEN];	/*Die einzelnen Parameter*/
	char						password[PASSWORD_LEN];
	char						userid_str[5];
	long int						userid;
	char						group_str[MAXLINE];
	long int						group;
	char						fullname[MAXLINE];
	char						shell[MAXLINE];
	char						action[32];
	entry						entries[MAXENTRIES];
	int						cl;
	int						x,m;
	FILE						*group_file;
	int						parameters;
	char						salt[2];					/*Fuer Passwort-Verschluesseln*/
   time_t					tm;						/*Time fuer Passwort*/

	error=0;

	/*debugging*/
	if (DEBUG) header(BENUTZER,"Debug");
   
   /*debugging*/	
   if (DEBUG) printf("Env-Variable lesen\n");
   
	/*Die Benutzereingaben einlesen*/
   env=getenv("CONTENT_LENGTH");
   cl = env?atoi(env):0;
	if (cl == 0) error=NO_INPUT_STRING;
	
	/*debugging*/
	if (DEBUG) printf("cl: %d, Error: %d<br>",cl,error);
	
	/*Hier werden die Daten eingelesen und in einem Feld des typs "entry" abgelegt*/

	if (!error){
    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
    }
   }
   parameters=x;
   
   /*debugging*/
	if (DEBUG) printf("parameters: %d<br>",parameters);
	   
	/*Die passwd-Datei einlesen*/
	if (!error) error=read_users(&user_list);
   
      
   /*Die group-Datei einlesen*/
 	if (!error) error=read_groups(&group_list);

	/*Die einzelnen Parameter herausnehmen und ueberpruefen*/
   /*action wird als erstes ausgewertet, da bei action=aktualisieren
     keine Defaultwerte ermittelt werden duerfen*/
   if (!error) {  
   	if (strcmp(entries[7].val,"")) strcpy(action,entries[7].val);
   	else error=NO_ACTION_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("Action: %s<br>",action);


	/*Der Username wird ausgewertet*/
   if (!error) {
		strcpy(username,entries[0].val);
		if (!strcmp(username,"")) error=NO_USERNAME_GIVEN;
   }
	/*debugging*/
	if (DEBUG) printf("username: %s<br>",username);

	if (!error) {
	   
	   /*Die UserId wird ausgewertet*/
		strcpy(userid_str,entries[1].val);
		if (strcmp(userid_str,"")){
			if (!isNumber(userid_str) && !error) error=USERID_NOT_NUMBER;
			if (!error) userid=atoi(userid_str);
		}
		else {
			if (strcmp(action,"aktualisieren")) userid=get_unused_uid(user_list);
         else userid=-1;
      }
      /*debugging*/
		if (DEBUG) printf("userid: %d<br>",userid);

		/*Die GroupId wird ausgewertet*/
		strcpy(group_str,entries[2].val);
		if (strcmp(group_str,"")){
			if (!isNumber(group_str) && !error) error=GROUP_NOT_NUMBER;
			if (!error) group=atoi(group_str);
		}
		else {
			if (strcmp(action,"aktualisieren")){
				if (!GroupnameUsed(group_list,username)) {
					group=get_unused_gid(group_list);
					new_group(group_list,group,username);
				}
				else group=getGroupId(group_list,username);
			}
			else group=-1;
		}
		/*debugging*/
		if (DEBUG) printf("group: %d<br>",group);


		/*Die Userinfo wird ausgewertet*/
		strcpy(fullname,entries[3].val);

		/*debugging*/
		if (DEBUG) printf("fullname: %s<br>",fullname);

		/*Das Passwort wird ausgewertet und codiert*/
		if (!strcmp(entries[4].val,entries[5].val)) {
			time(&tm);
			salt[0] = (tm & 0x0f) +	'A';
			salt[1] = ((tm & 0xf0) >> 4) + 'a';
 			if (strlen(entries[4].val) > 0) strcpy(password,crypt(entries[4].val,salt));
			else strcpy(password,"");
			/*debugging*/
			if (DEBUG) printf("Password: %s<br>",password);
		}
		else error=PASSWORDS_DIFFERENT;


		/*Die Shell wird ausgewertet*/
		strcpy(shell,entries[6].val);
		if ((!strcmp(shell,"default")) && (strcmp(action,"aktualisieren"))) strcpy(shell,DEFAULTSHELL);
	
		/*debugging*/
		if (DEBUG) printf("entries[6].name: %s, entries[6].val: %s, shell: %s<br>",entries[6].name,entries[6].val,shell);

	}	
	
	/*debugging*/
	if (DEBUG) printf("Vor der Aktion: Error: %d<br>",error);
	
		
   /*Hier wird ausgewaehlt, was der User will*/
   if (!error){
		if (!strcmp(action,"eintragen")) { 
			if (DEBUG) printf("main: add_user aufrufen<br>");
			error=add_user(user_list, group_list, username, password, userid, group, fullname, shell);
		}
		else { 
			if (!strcmp(action,"entfernen") && !error) {
				if (DEBUG) printf("main: delete_user aufrufen<br>");
				error=delete_user(user_list, group_list, username, password, userid, group, fullname, shell);
			}
			else { 
				if (DEBUG) printf("main: change_user aufrufen<br>");
				if (!strcmp(action,"aktualisieren") && !error) error=change_user(user_list, group_list, username, password, userid, group, fullname, shell);
			}
		}				
   }
   

	/*debugging*/
	printf("Nach der Aktion: Error: %d<br>",error);
	if (DEBUG) tail();

	if (!DEBUG) {
		if (error == NO_ERROR) {
			header(NONE,"Erfolgreich");
			printf("Die Aktion wurde erfolgreich abgeschlossen");
			tail();
		}
		else
			{
				switch (error) {
					case QUERY_STRING_NOT_SET : print_error("Es wurden keine Parameter empfangen ($QUERY_STRING nicht gesetzt).");
					     break;
					case USERID_NOT_NUMBER : print_error("Die Benutzernummer ist keine Zahl");
					     break;
					case GROUP_NOT_NUMBER : print_error("Die Gruppennummer ist keine Zahl");
					     break;
					case USER_EXISTS : print_error("Der Benutzer existiert bereits");
				   	  break;
					case USER_UNKNOWN : print_error("Der Benutzer ist unbekannt");
					     break;
					case USERID_EXISTS : print_error("Die Benutzernummer ist bereits vergeben");
					     break;
					case USERID_UNKNOWN : print_error("Die Benutzernummer ist unbekannt");
				   	  break;
					case GROUP_EXISTS : print_error("Die Gruppe existiert bereits");
					     break;
					case GROUP_UNKNOWN : print_error("Die Gruppe ist unbekannt");
					     break;
					case PASSWD_FAILURE : print_error("Die Datei /etc/passwd konnte nicht gelesen/geschrieben werden");
				   	  break;
					case GROUP_FAILURE : print_error("Die Datei /etc/group konnte nicht gelesen/geschrieben werden");
					     break;
					case NO_USERNAME_GIVEN : print_error("Sie m&uuml;ssen einen Benutzernamen angeben!");
					     break;
					case NO_ACTION_GIVEN : print_error("Sie m&uuml;ssen hinzuf&uuml;gen, l&ouml;schen oder &auml;ndern ausw&auml;hlen!");
				   	  break;
					case NO_INPUT_STRING : print_error("Es wurden keine Eingaben aus dem Browser empfangen!");
					     break;
					case PASSWORDS_DIFFERENT : print_error("Die eingegebenen Passw&ouml;rter stimmen nicht &uuml;berein.");
					     break;
				}
			}
	}
}
