/*
 * lpspool.c: looks "/pipe/<printername>" pipes. It is a program you can use
 * to do remote printing with NVDI. All data you copy into /pipe/<printername> is piped
 * to the corresponding printer of the lpr system. This is copyrighted software; see the
 * file "COPYING" for details.
 * lpspool.c is based on the lpdev written by Thierry Bousch.
 *
 * This file must be compiled with 16-bit integers.
 *
 * Author: Torsten Lang, Entw.-Bro f. Hard- u. Software
 * Version: 0.1 (September 1997)
 *
 * Revision history:
 *  0.1: First attempt.
 *  0.2: Version for MiNT'98 V1.0
 *  0.3: Changes for NVDI5
 */

#include "lpspool.h"
#define  LP_VERSION	"0.3"

/*
 * Global variables
 */

struct printerinfo printer[MAX_PRINTERS];
int printers = 0;

/* Copyright information */

void Version (void)
{
	Cconws("Spooler driver for lpr,  1997 by T. Lang, Entw.-Bro f. Hard- u. Software\r\n"
	       "(version " LP_VERSION ").\r\n\r\n");
}


/* 
 * Installs everything, returns 0 on success. Must be executed in
 * supervisor mode.
 */

long install_things (void)
{
	FILE *file;
	char *buffer = "U:\\pipe\\                                                                ";
	int j;

	if (Syield() == EINVFN) {
		Cconws("lpspool: MiNT is not running\r\n");
		return EACCDN;
	}

	/* open the /etc/printcap file for reading */
	if (!(file = fopen("U:\\etc\\printcap", "r")))
		if (!(file = fopen("C:\\etc\\printcap", "r"))) {
			Cconws("lpspool: no /etc/printcap file, device not installed\r\n\r\n");
			return EACCDN;
	}
	
	while (printers < MAX_PRINTERS) {
		/* extract a printer's name from /etc/printcap */
		buffer[70] = 0;
	
		for (j = 8; j < 70; j++) {
			if ((buffer[j] = fgetc(file)) == EOF)
				break;
			if ((buffer[j] == 0x20) || (buffer[j] == 0x09)) {
				j--;
				continue;
			}
			if ((buffer[j] == 0x3A) || (buffer[j] == 0x7C)) {
				again:
				while((buffer[j+2] = fgetc(file)) != 0x0A) {
					if (buffer[j+2] == EOF)
						break;
					if ((buffer[j+2] == 0x20) || (buffer[j+2] == 0x09))
						continue;
					buffer[j] = buffer[j+1];
					buffer[j+1] = buffer[j+2];
				}
				if ((buffer[j+1] != 0x3A) || (buffer[j+2] != 0x0A))
					goto again;
				buffer[j] = 0;
				break;
			}
		}
		if (buffer[8] == EOF)
			break;

		strcpy(printer[printers].name, buffer);

		Cconws("lpspool: spooler for /pipe/");
		Cconws(&buffer[8]);
		Cconws(" installed\r\n");
		
		printers++;
		if ((buffer[j] == EOF) || (buffer[j+2] == EOF))
			break;
	}
		
	fclose(file);
	
	return 0;
}

void alrm(void)
{
}

/*
 * The main routine is very simple now, it just calls install_things
 * and remains resident if everything went well
 */

int main()
{
	long ret;
	FILE *file;
	int i;
	char buffer[128];
	
	Version();

	ret = Supexec(install_things);
	if (ret < 0)
		return ret;

	Psignal(SIGALRM, alrm);
	/* Installation is complete */
	while(1) {
		for (i = 0; i < printers; i++) {
			if ((file = fopen(printer[i].name, "rb"))) {
				fclose(file);
				sprintf(buffer, "sh -c \"/usr/bin/cat /pipe/%s | /usr/ucb/lpr -P %s -l -\"", (char *)&printer[i].name[8], (char *)&printer[i].name[8]);
				system(buffer);
				while ((file = fopen(printer[i].name, "rb"))) {
					fclose(file);
					Talarm(5L);
					Psigpause(0L);
				}
			}
		}
		Talarm(5L);
		Psigpause(0L);
	}
	return -999;	/* never reached, just to make Gcc happy */
}
