/*
 * lpspool.c: installs "/dev/<printername>" devices. It is a device you can use
 * to do remote printing with NVDI. All data you copy into /dev/<printername> is piped
 * to the corresponding printer of the lpr system. This is copyrighted software; see the
 * file "COPYING" for details.
 * lpspool.c is based on the lpdev written by Thierry Bousch.
 *
 * This file must be compiled with 16-bit integers.
 *
 * Author: Torsten Lang, Entw.-Bro f. Hard- u. Software
 * Version: 0.1 (September 1997)
 *
 * Revision history:
 *  0.1: First attempt.
 */

#include "lpspool.h"
#define  LP_VERSION	"0.1"

/*
 * Global variables
 */
 
struct kerinfo *kernel;
struct printerinfo printer[MAX_PRINTERS];
int printers = 0;

/*
 * Forward declarations of the device driver functions
 */

long	lpspool_open		(FILEPTR *f);
long	lpspool_write		(FILEPTR *f, char *buf, long bytes);
long	lpspool_read		(FILEPTR *f, char *buf, long bytes);
long	lpspool_lseek		(FILEPTR *f, long where, int whence);
long	lpspool_ioctl		(FILEPTR *f, int mode, void *buf);
long	lpspool_datime		(FILEPTR *f, int *timeptr, int rwflag);
long	lpspool_close		(FILEPTR *f, int pid);
long	lpspool_select		(FILEPTR *f, long proc, int mode);
void	lpspool_unselect	(FILEPTR *f, long proc, int mode);

/* Copyright information */

void Version (void)
{
#ifdef XDD
	Cconws("Device driver for lpr,  1997 by T. Lang, Entw.-Bro f. Hard- u. Software\r\n"
	       "(version " LP_VERSION ").\r\n\r\n");
#else
	Cconws("Device driver for lpr,  1997 by T. Lang, Entw.-Bro f. Hard- u. Software\r\n"
	       "(version " LP_VERSION ").\r\n\r\n");
#endif
}

/* 
 * Installs everything, returns 0 on success. Must be executed in
 * supervisor mode.
 */

long install_things (void)
{
	FILE *file;
	char *buffer = "U:\\dev\\                                                                 ";
	int j;

#ifndef XDD
	if (Syield() == EINVFN) {
		Cconws("lpspool: MiNT is not running\r\n");
		return EACCDN;
	}
#endif
	/* open the /etc/printcap file for reading */
	if (!(file = fopen("U:\\etc\\printcap", "r")))
		if (!(file = fopen("C:\\etc\\printcap", "r"))) {
			Cconws("lpspool: no /etc/printcap file, device not installed\r\n\r\n");
			return EACCDN;
	}
	
	while (printers < MAX_PRINTERS) {
		/* extract a printer's name from /etc/printcap */
		buffer[70] = 0;
	
		for (j = 7; j < 70; j++) {
			if ((buffer[j] = fgetc(file)) == EOF)
				break;
			if ((buffer[j] == 0x20) || (buffer[j] == 0x09)) {
				j--;
				continue;
			}
			if ((buffer[j] == 0x3A) || (buffer[j] == 0x7C)) {
				again:
				while((buffer[j+2] = fgetc(file)) != 0x0A) {
					if (buffer[j+2] == EOF)
						break;
					if ((buffer[j+2] == 0x20) || (buffer[j+2] == 0x09))
						continue;
					buffer[j] = buffer[j+1];
					buffer[j+1] = buffer[j+2];
				}
				if ((buffer[j+1] != 0x3A) || (buffer[j+2] != 0x0A))
					goto again;
				buffer[j] = 0;
				break;
			}
		}
		if (buffer[7] == EOF)
			break;

		printer[printers].lpspool_device.open     = lpspool_open;
		printer[printers].lpspool_device.write    = lpspool_write;
		printer[printers].lpspool_device.read     = lpspool_read;
		printer[printers].lpspool_device.lseek    = lpspool_lseek;
		printer[printers].lpspool_device.ioctl    = lpspool_ioctl;
		printer[printers].lpspool_device.datime   = lpspool_datime;
		printer[printers].lpspool_device.close    = lpspool_close;
		printer[printers].lpspool_device.select   = lpspool_select;
		printer[printers].lpspool_device.unselect = lpspool_unselect;
		printer[printers].our_lock.l_type   = F_WRLCK;
		printer[printers].our_lock.l_whence =  0;
		printer[printers].our_lock.l_start  =  0L;
		printer[printers].our_lock.l_len    =  0L;
		printer[printers].our_lock.l_pid    = -1;
		printer[printers].open = 0;
		printer[printers].devinfo.driver = &printer[printers].lpspool_device;
		printer[printers].selector = 0L;
       	
#ifndef XDD
		if (Fsfirst(buffer, 0) == 0) {
			Cconws("lpspool: device /dev/");
			Cconws(&buffer[7]);
			Cconws(" already installed\r\n");
			continue;
		}
#endif
		kernel = (struct kerinfo *)Dcntl(DEV_INSTALL, buffer, &printer[printers].devinfo);
		if ((long)kernel <= 0L) {
			Cconws("lpspool: unable to install device\r\n\r\n");
			return EACCDN;
		}
		else {
			Cconws("lpspool: device /dev/");
			Cconws(&buffer[7]);
			Cconws(" installed\r\n");
		}
		
		printers++;
		if ((buffer[j] == EOF) || (buffer[j+2] == EOF))
			break;
	}
		
	fclose(file);
	
	return 0;
}

#ifndef XDD
/*
 * The main routine is very simple now, it just calls install_things
 * and remains resident if everything went well
 */

int main()
{
	long ret;
	
	Version();

	ret = Supexec(install_things);
	if (ret < 0)
		return ret;

	/* Installation is complete */
	Ptermres(256L + _base->p_tlen + _base->p_dlen + _base->p_blen, 0);
	return -999;	/* never reached, just to make Gcc happy */
}

#else

/* uk: if this is linked and started as an externel device driver,
 *     the memory allocation stuff is already done and we are in
 *     super mode already.
 */

DEVDRV * xdd_main(struct kerinfo *k)
{
	long res;

	kernel = k;
	Version();
	res = install_things();
	if (res == 0)
		return (DEVDRV*)1;
	else
		return NULL;
}

#endif

/*
 * Will wake any process select'ing the printer;
 * this routine is called by the interrupt handler, but also when the
 * buffer is flushed.
 */

void wake_up (void)
{
	int i;
	
	for (i = 0; i < printers; i++)
		if (printer[i].selector)
			WAKESELECT(printer[i].selector);	/* wake selector */
}

/*
 * Here are the actual device driver functions
 */
 
#define  LP_LOCKED(i)	(printer[i].our_lock.l_pid >= 0)

long lpspool_open (FILEPTR *f)
{
	int i;
	char buffer[32];
	
	for (i = 0; i < printers; i++)
		if (f->dev == &printer[i].lpspool_device)
			break;

	if (printer[i].open)
		return EACCDN;

	printer[i].open = 1;
	
	/* open pipe for output */
	sprintf(buffer, "lpr -h -l U:\\pipe\\lpspool.%03d", i);
	printer[i].file = fopen(&buffer[10], "wb");
	buffer[29] = 0x20;
	buffer[30] = 0x26;
	buffer[31] = 0;
	system(buffer);
	TRACE(("lpspool: open device"));
	return 0;
}

long lpspool_close (FILEPTR *f, int pid)
{
	int i;

	for (i = 0; i < printers; i++)
		if (f->dev == &printer[i].lpspool_device)
			break;

	if (!printer[i].open)
		return EACCDN;

	printer[i].open = 0;

	/* close pipe */
	fclose(printer[i].file);
	
	TRACE(("lpspool: close device"));
	if ((f->flags & O_LOCK) && printer[i].our_lock.l_pid == pid) {
		TRACE(("lpspool: releasing lock on close"));
		f->flags &= ~O_LOCK;
		printer[i].our_lock.l_pid = -1;
		WAKE(IO_Q, (long)&printer[i].our_lock);
	}
	return 0;
}

long lpspool_read (FILEPTR *f, char *buf, long bytes)
{
	TRACE(("lpspool: foolish attempt to read"));
	return 0;
}

long lpspool_datime (FILEPTR *f, int *timeptr, int rwflag)
{
	if (rwflag) {
		DEBUG(("lpspool: can't modify date/time"));
		return EACCDN;
	}
	TRACE(("lpspool: read time and date"));
	*timeptr++ = TGETTIME();
	*timeptr   = TGETDATE();
	return 0;
}

long lpspool_lseek (FILEPTR *f, long where, int whence)
{
	TRACE(("lpspool: foolish attempt to seek"));
	if (whence < 0 || whence > 2)
		return EINVFN;
	return where ? ERANGE : 0L;
}

long lpspool_ioctl (FILEPTR *f, int mode, void *buf)
{
	struct flock *g;
	int i;
	
	for (i = 0; i < printers; i++)
		if (f->dev == &printer[i].lpspool_device)
			break;

	if (mode == FIONREAD) {
		TRACE(("lpspool: ioctl(FIONREAD)"));
	}
	else if (mode == FIONWRITE) {
		TRACE(("lp: ioctl(FIONWRITE)"));
	}
	else if (mode == TIOCFLUSH) {
		TRACE(("lp: clear buffer"));
		wake_up();	/* Wake up any select'ing process */
	}
	else if (mode == F_GETLK) {
		g = (struct flock *) buf;

		if (LP_LOCKED(i)) {
			TRACE(("lpspool: get_lock succeeded"));
			*g = printer[i].our_lock;
		} else {
			TRACE(("lpspool: get_lock failed"));
			g->l_type = F_UNLCK;
		}
	}
	else if (mode == F_SETLK || mode == F_SETLKW) {
		g = (struct flock *) buf;

		switch (g->l_type) {
		case F_UNLCK:
		    if (!(f->flags & O_LOCK) || g->l_pid != printer[i].our_lock.l_pid) {
			DEBUG(("lpspool: no such lock"));
			return ENSLOCK;
		    } else {
			TRACE(("lpspool: remove lock"));
			f->flags &= ~O_LOCK;
			printer[i].our_lock.l_pid = -1;
			WAKE(IO_Q, (long)&printer[i].our_lock);
		    }
		    return 0;
		case F_RDLCK:
		    TRACE(("lpspool: read locks are ignored"));
		    return 0;
		case F_WRLCK:
		    while (LP_LOCKED(i)) {
			DEBUG(("lpspool: conflicting locks"));
			if (mode == F_SETLK) {
				*g = printer[i].our_lock;
				return ELOCKED;
			}
			SLEEP(IO_Q, (long)&printer[i].our_lock);
		    }
		    TRACE(("lpspool: set lock"));
		    f->flags |= O_LOCK;
		    printer[i].our_lock.l_pid = g->l_pid;
		    return 0;
		default:
		    DEBUG(("lpspool: invalid lock type"));
		    return EINVFN;
		}
	}
	else {
		DEBUG(("lp: invalid ioctl mode"));
		return EINVFN;
	}
	return 0;
}

long lpspool_write (FILEPTR *f, char *buf, long bytes)
{
	int i;
	long count;
	long total = bytes;
	
	for (i = 0; i < printers; i++)
		if (f->dev == &printer[i].lpspool_device)
			break;

	while (bytes) {
		count = fwrite(buf, 1, bytes, printer[i].file);
		bytes -= count;
		buf += count;
	}

	return total;
}

/* Bug: only one process can select the printer */

long lpspool_select (FILEPTR *f, long proc, int mode)
{
	int i;
	
	for (i = 0; i < printers; i++)
		if (f->dev == &printer[i].lpspool_device)
			break;

	if (!printer[i].selector) {
		TRACE(("lpspool: select returned 0"));
		printer[i].selector = proc;
		return 0;
	}
	TRACE(("lpspool: select returned 1"));
	return 1;
}

void lpspool_unselect (FILEPTR *f, long proc, int mode)
{
	int i;
	
	for (i = 0; i < printers; i++)
		if (f->dev == &printer[i].lpspool_device)
			break;

	TRACE(("lpspool: unselect"));
	printer[i].selector = 0L;
}
